import typing as t
import uuid

import httpx
from openai import OpenAI
from openai._exceptions import APIStatusError
from openai._models import FinalRequestOptions
from openai._types import Omit
from openai._utils import is_given
import sempy.fabric as fabric
from sempy.fabric._token_provider import SynapseTokenProvider
from sempy.fabric.exceptions import FabricHTTPException
from ._util import (
    get_artifact_by_id_or_name,
    get_workspace_capacity_id,
    resolve_workspace_name_and_id,
)
from synapse.ml.fabric.service_discovery import get_fabric_env_config


MONIKER_HEADER_NAME = "x-ms-workload-resource-moniker"
ARTIFACT_HEADER_NAME = "x-ms-upstream-artifact-id"
BYPASS_COMSOS_DB_HEADER_NAME = "x-ms-ai-assistant-bypass-cosmosdb"
FABRIC_RUN_ID_HEADER_NAME = "x-ms-ai-assistant-fabric-run-id"
ASSISTANT_SCENARIO_HEADER_NAME = "x-ms-ai-assistant-scenario"
AI_SKILL_STAGE_HEADER_NAME = "x-ms-ai-aiskill-stage"

AI_SKILL_STAGE_T = t.Literal["sandbox", "production", None]
ASSISTANT_SCENARIOS_T = t.Literal["aiskill", "dscopilot", "dwcopilot", "aiskillv1shim"]
FEATURE_NAMES_T: t.TypeAlias = t.Literal["AISkill", "DSCopilot", "DWCopilot"]


class FabricOpenAI(OpenAI):
    """
    Interact with Fabric's AI Assistant API.

    Parameters
    ----------
    artifact_name : str
        The name of the Data Agent artifact.
    workspace_name : str, optional
        The name of the workspace. If not provided, it defaults to the notebook's workspace name.
    api_version : str, default="2024-05-01-preview"
        The API version to be used for the OpenAI.
    assistant_scenario : str, default="aiskill"
        The scenario for the assistant. Options are "aiskill", "dscopilot", "dwcopilot", "aiskillv1shim".
    ai_skill_stage : str, default="production"
        The stage of the Data Agent. Options are "sandbox", "production".
    kwargs : Any
        Additional keyword arguments.
    """

    def __init__(
        self,
        artifact_name: str,
        workspace_name: str | None = None,
        api_version: str = "2024-05-01-preview",
        assistant_scenario: ASSISTANT_SCENARIOS_T = "aiskill",
        ai_skill_stage: AI_SKILL_STAGE_T = "sandbox",
        **kwargs: t.Any,
    ) -> None:
        """
        Initialize a FabricOpenAI object.

        Parameters
        ----------
        artifact_name : str
            The name of the Data Agent artifact.
        workspace_name : str, optional
            The name of the workspace. If not provided, it defaults to the notebook's workspace name.
        api_version : str, default="2024-05-01-preview"
            The API version to be used for the OpenAI.
        assistant_scenario : str, default="aiskill"
            The scenario for the assistant. Options are "aiskill", "dscopilot", "dwcopilot", "aiskillv1shim".
        ai_skill_stage : str, default="sandbox"
            The stage of the Data Agent. Options are "sandbox", "production".
        kwargs : Any
            Additional keyword arguments.
        """
        self.workspace_name, self.workspace_id = resolve_workspace_name_and_id(
            workspace_name
        )
        self.capacity_id = get_workspace_capacity_id(self.workspace_id)

        self.artifact_name, self.artifact_id = get_artifact_by_id_or_name(
            artifact_name, ["DataAgent", "AISkill", "Warehouse"], self.workspace_id
        )

        self.api_version = api_version
        self.assistant_scenario = assistant_scenario
        self.ai_skill_stage = ai_skill_stage
        default_query = kwargs.pop("default_query", {})
        default_query["api-version"] = self.api_version
        self._moniker_id = str(uuid.uuid4())

        self._host = get_fabric_env_config().fabric_env_config.wl_host
        # TODO: Change when base_url is modified
        base_url = f"{self._host}/webapi/capacities/{self.capacity_id}/workloads/ML/AISkill/Automatic/v1/workspaces/{self.workspace_id}/artifacts/{self.artifact_id}/aiassistant/openai"

        super().__init__(
            api_key="",
            base_url=base_url,
            default_query=default_query,
            **kwargs,
        )

    def _prepare_options(self, options: FinalRequestOptions) -> FinalRequestOptions:
        """
        Prepare the request options by setting headers and other necessary parameters.

        Parameters
        ----------
        options : FinalRequestOptions
            The request options to be prepared.
        """
        headers: dict[str, str | Omit] = (
            {**options.headers} if is_given(options.headers) else {}
        )
        options.headers = headers
        headers["Authorization"] = f"Bearer {SynapseTokenProvider()()}"
        if "Accept" not in headers:
            headers["Accept"] = "application/json"
        if "ActivityId" not in headers:
            correlation_id = str(uuid.uuid4())
            headers["ActivityId"] = correlation_id
        if MONIKER_HEADER_NAME not in headers:
            headers[MONIKER_HEADER_NAME] = self._moniker_id
        headers[ASSISTANT_SCENARIO_HEADER_NAME] = self.assistant_scenario
        if self.assistant_scenario == "aiskill":
            headers[AI_SKILL_STAGE_HEADER_NAME] = self.ai_skill_stage

        return super()._prepare_options(options)

    def _make_status_error(
        self, err_msg: str, *, body: object, response: httpx.Response
    ) -> APIStatusError:
        """
        Create and return an API status error with additional information from the response headers.

        Parameters
        ----------
        err_msg : str
            The error message.
        body : object
            The body of the response.
        response : httpx.Response
            The HTTP response object.

        Returns
        -------
        APIStatusError
            The API status error with additional information such as RAID and correlation ID.
        """
        if raid := response.headers.get("x-ms-root-activity-id", None):
            err_msg += f" RAID: {raid}"
        if correlation_id := response.headers.get("ActivityId", None):
            err_msg += f" Correlation ID: {correlation_id}"

        return super()._make_status_error(err_msg, body=body, response=response)

    def get_or_create_thread(self, tag: str = ""):
        """
        Retrieve or create a thread with the specified tag.

        Parameters
        ----------
        tag : str
            The tag for the thread. No tag means get the default thread.

        Returns
        -------
        UUID
            The ID of the thread.
        """
        headers = {
            "Content-Type": "application/json",
            "x-ms-workload-resource-moniker": f"{self._moniker_id}",
        }

        artifact_url = f"{self._host}/webapi/capacities/{self.capacity_id}/workloads/ML/AISkill/Automatic/v1/workspaces/{self.workspace_id}/artifacts/{self.artifact_id}"
        client = fabric.FabricRestClient()

        response = client.get(
            f"{artifact_url}/aiassistant/threads/fabric?agent=aiskill&stage={self.ai_skill_stage}&client=dataagentsdk&tag={tag}&scenario={self.assistant_scenario}",
            headers=headers,
        )

        if response.status_code != 200:
            raise FabricHTTPException(response)

        return self.beta.threads.retrieve(response.json()['id'])

    def delete_thread(self, tag: str = ""):
        """
        Delete a thread with the specified tag.

        Parameters
        ----------
        tag : str
            The tag for the thread. No tag means delete the default thread.
        """
        headers = {
            "Content-Type": "application/json",
            "x-ms-workload-resource-moniker": f"{self._moniker_id}",
        }

        artifact_url = f"{self._host}/webapi/capacities/{self.capacity_id}/workloads/ML/AISkill/Automatic/v1/workspaces/{self.workspace_id}/artifacts/{self.artifact_id}"
        client = fabric.FabricRestClient()

        response = client.delete(
            f"{artifact_url}/aiassistant/threads/fabric?agent=aiskill&stage={self.ai_skill_stage}&client=dataagentsdk&tag={tag}&scenario={self.assistant_scenario}",
            headers=headers,
        )

        if response.status_code != 200:
            raise FabricHTTPException(response)



from fabric.dataagent.client._fabric_openai import FabricOpenAI
from fabric.dataagent.client._fabric_data_agent_api import FabricDataAgentAPI
from fabric.dataagent.client._create_delete import create_data_agent, delete_data_agent
from fabric.dataagent.client._fabric_data_agent_mgmt import FabricDataAgentManagement
from fabric.dataagent.client._datasource import Datasource
from fabric.dataagent.client._tagged_value import TaggedValue

__all__ = [
    "Datasource",
    "FabricOpenAI",
    "FabricDataAgentAPI",
    "FabricDataAgentManagement",
    "TaggedValue",
    "create_data_agent",
    "delete_data_agent",
]



import unittest
from unittest.mock import MagicMock, patch
from uuid import UUID
from fabric.dataagent.client._fabric_data_agent_api import FabricDataAgentAPI
from fabric.dataagent.client._tagged_value import TaggedValue


class TestFabricDataAgentAPI(unittest.TestCase):

    @patch('fabric.dataagent.client._fabric_data_agent_api.get_artifact_by_id_or_name')
    @patch('fabric.dataagent.client._fabric_data_agent_api.get_workspace_capacity_id')
    @patch(
        'fabric.dataagent.client._fabric_data_agent_api.resolve_workspace_name_and_id'
    )
    @patch('synapse.ml.fabric.service_discovery.get_fabric_env_config')
    @patch('sempy.fabric.FabricRestClient')
    def setUp(
        self,
        MockFabricRestClient,
        MockGetFabricEnvConfig,
        MockResolveWorkspaceNameAndID,
        MockGetWorkspaceCapacityID,
        MockGetArtifactByIDOrName,
    ):
        self.mock_client = MockFabricRestClient.return_value
        self.mock_get_fabric_env_config = MockGetFabricEnvConfig
        self.mock_resolve_workspace_name_and_id = MockResolveWorkspaceNameAndID
        self.mock_get_workspace_capacity_id = MockGetWorkspaceCapacityID
        self.mock_get_artifact_by_id_or_name = MockGetArtifactByIDOrName

        self.mock_resolve_workspace_name_and_id.return_value = (
            "workspace_name",
            UUID("12345678123456781234567812345678"),
        )
        self.mock_get_workspace_capacity_id.return_value = "capacity_id"
        self.mock_get_artifact_by_id_or_name.return_value = (
            "data_agent_name",
            UUID("12345678123456781234567812345678"),
        )
        self.mock_get_fabric_env_config.return_value.fabric_env_config.wl_host = (
            "http://fabric_host"
        )

        self.data_agent_api = FabricDataAgentAPI(
            data_agent="test_data_agent", workspace="test_workspace"
        )

    def test_get_configuration(self):
        self.mock_client.get.return_value = MagicMock(
            status_code=200, json=lambda: {}, headers={"ETag": "etag"}
        )
        config = self.data_agent_api.get_configuration()
        self.assertIsInstance(config, TaggedValue)
        self.assertEqual(config.etag, "etag")

    def test_set_configuration(self):
        self.mock_client.patch.return_value = MagicMock(
            status_code=200, headers={"ETag": "new_etag"}
        )
        value = TaggedValue({"key": "value"}, "etag")
        updated_config = self.data_agent_api.set_configuration(value)
        self.assertIsInstance(updated_config, TaggedValue)
        self.assertEqual(updated_config.etag, "new_etag")

    def test_get_datasources_empty(self):
        self.mock_client.get.side_effect = [
            MagicMock(
                status_code=200,
                json=lambda: {},
            )
        ]
        datasources = list(self.data_agent_api.get_datasources())
        self.assertEqual(len(datasources), 0)

    def test_get_datasources(self):
        self.mock_client.get.side_effect = [
            MagicMock(
                status_code=200,
                json=lambda: {"dataSources": [{"id": "ds1"}, {"id": "ds2"}]},
            ),
            MagicMock(
                status_code=200, json=lambda: {"id": "ds1"}, headers={"ETag": "etag"}
            ),
            MagicMock(
                status_code=200, json=lambda: {"id": "ds2"}, headers={"ETag": "etag"}
            ),
        ]
        datasources = list(self.data_agent_api.get_datasources())
        self.assertEqual(len(datasources), 2)
        self.assertEqual(datasources[0].value["id"], "ds1")

    def test_get_datasource(self):
        self.mock_client.get.return_value = MagicMock(
            status_code=200, json=lambda: {}, headers={"ETag": "etag"}
        )
        datasource = self.data_agent_api.get_datasource("ds1")
        self.assertIsInstance(datasource, TaggedValue)
        self.assertEqual(datasource.etag, "etag")

    def test_remove_datasource(self):
        self.mock_client.delete.return_value = MagicMock(status_code=200)
        self.data_agent_api.remove_datasource("ds1", "etag")
        self.mock_client.delete.assert_called_once()

    def test_set_datasource(self):
        self.mock_client.put.return_value = MagicMock(
            status_code=200, headers={"ETag": "new_etag"}
        )
        value = TaggedValue({"id": "ds1"}, "etag")
        updated_datasource = self.data_agent_api.set_datasource(value)
        self.assertIsInstance(updated_datasource, TaggedValue)
        self.assertEqual(updated_datasource.etag, "new_etag")

    def test_get_publishing_info(self):
        self.mock_client.get.return_value = MagicMock(
            status_code=200, json=lambda: {}, headers={"ETag": "etag"}
        )
        publishing_info = self.data_agent_api.get_publishing_info()
        self.assertIsInstance(publishing_info, TaggedValue)
        self.assertEqual(publishing_info.etag, "etag")

    def test_put_publish_info(self):
        self.mock_client.put.return_value = MagicMock(
            status_code=200, json=lambda: {}, headers={"ETag": "new_etag"}
        )
        value = TaggedValue({"description": "value"}, "etag")
        updated_publishing_info = self.data_agent_api.put_publish_info(value)
        self.assertIsInstance(updated_publishing_info, TaggedValue)
        self.assertEqual(updated_publishing_info.etag, "new_etag")

    def test_add_datasource(self):
        self.mock_client.post.return_value = MagicMock(
            status_code=200, json=lambda: {"id": "new_ds"}
        )
        new_datasource = self.data_agent_api.add_datasource({"key": "value"})
        self.assertEqual(new_datasource["id"], "new_ds")

    def test_publish(self):
        self.mock_client.put.return_value = MagicMock(status_code=200)
        value = TaggedValue({}, "etag")
        self.data_agent_api.publish(value)
        self.mock_client.put.assert_called_once()

    def test_get_schema(self):
        self.mock_client.get.return_value = MagicMock(
            status_code=200, json=lambda: {"schema": {}}
        )
        schema = self.data_agent_api.get_schema("workspace_id", "datasource_id", "type")
        self.assertEqual(schema, {"schema": {}})

    def test_get_datasource_fewshots(self):
        self.mock_client.get.return_value = MagicMock(
            status_code=200, json=lambda: {}, headers={"ETag": "etag"}
        )
        fewshots = self.data_agent_api.get_datasource_fewshots("ds1")
        self.assertIsInstance(fewshots, TaggedValue)
        self.assertEqual(fewshots.etag, "etag")

    def test_set_datasource_fewshots(self):
        self.mock_client.post.return_value = MagicMock(status_code=200)
        value = TaggedValue({"fewShots": []}, "etag")
        self.data_agent_api.set_datasource_fewshots("ds1", value)
        self.mock_client.post.assert_called_once()



import unittest
from unittest.mock import patch

from fabric.dataagent.client._datasource import Datasource
from fabric.dataagent.client._fabric_data_agent_mgmt import FabricDataAgentManagement
from fabric.dataagent.client._tagged_value import TaggedValue


class TestFabricDataAgentManagement(unittest.TestCase):

    @patch('fabric.dataagent.client._fabric_data_agent_mgmt.FabricDataAgentAPI')
    @patch('fabric.dataagent.client._fabric_data_agent_mgmt.get_artifact_by_id_or_name')
    @patch('fabric.dataagent.client._fabric_data_agent_mgmt.list_items')
    @patch('fabric.dataagent.client._fabric_data_agent_mgmt.get_notebook_workspace_id')
    @patch('fabric.dataagent.client._fabric_data_agent_mgmt.resolve_workspace_id')
    def setUp(
        self,
        MockResolveWorkspaceID,
        MockGetNotebookWorkspaceID,
        MockListItems,
        MockGetArtifactByIDOrName,
        MockFabricDataAgentAPI,
    ):
        self.mock_client = MockFabricDataAgentAPI.return_value
        self.mock_get_artifact_by_id_or_name = MockGetArtifactByIDOrName
        self.mock_list_items = MockListItems
        self.mock_get_notebook_workspace_id = MockGetNotebookWorkspaceID
        self.mock_resolve_workspace_id = MockResolveWorkspaceID
        self.data_agent_management = FabricDataAgentManagement(
            data_agent="test_data_agent", workspace="test_workspace"
        )

    def test_update_configuration(self):
        self.mock_client.get_configuration.return_value = TaggedValue(
            {"additionalInstructions": "", "userDescription": ""}, "etag"
        )
        self.data_agent_management.update_configuration(
            instructions="New instructions", user_description="New description"
        )
        self.mock_client.set_configuration.assert_called_once()
        updated_config = self.mock_client.set_configuration.call_args[0][0]
        self.assertEqual(
            updated_config.value["additionalInstructions"], "New instructions"
        )
        self.assertEqual(updated_config.value["userDescription"], "New description")

    def test_get_configuration(self):
        self.mock_client.get_configuration.return_value = TaggedValue(
            {
                "additionalInstructions": "instructions",
                "userDescription": "description",
            },
            "etag",
        )
        config = self.data_agent_management.get_configuration()
        self.assertEqual(config.instructions, "instructions")
        self.assertEqual(config.user_description, "description")

    def test_publish(self):
        self.mock_client.get_configuration.return_value = TaggedValue({}, "etag")
        self.data_agent_management.publish()
        self.mock_client.publish.assert_called_once()

    def test_get_datasources(self):
        self.mock_client.get_datasources.return_value = iter(
            [TaggedValue({"id": "ds1"}, "etag"), TaggedValue({"id": "ds2"}, "etag")]
        )
        datasources = self.data_agent_management.get_datasources()
        self.assertEqual(len(datasources), 2)
        self.assertIsInstance(datasources[0], Datasource)
        self.assertEqual(datasources[0]._id, "ds1")

    def test_remove_datasource(self):
        self.mock_client.get_configuration.return_value = TaggedValue({}, "etag")
        self.mock_client.get_datasources.return_value = iter(
            [TaggedValue({"id": "ds1", "display_name": "datasource1"}, "etag")]
        )
        self.data_agent_management.remove_datasource("datasource1")
        self.mock_client.remove_datasource.assert_called_once_with("ds1", "etag")



def pytest_configure(config):
    """Early initialization before imports."""
    import os

    os.environ["REPORT_USAGE_TELEMETRY"] = "false"
    os.environ["fabric_fake_usage_telemetry"] = "true"


from pytest import fixture
import uuid
from .auth import get_fabric_info


@fixture(scope="session")
def fabric_info():
    """Get the authentication context for Fabric"""
    return get_fabric_info()


@fixture(scope="session")
def fabric_named_workspace(fabric_info):
    """Get or creates a workspace (dev name or fixed for CI/CD) and the authentication context for Fabric"""
    from sempy.fabric._client import import_pbix_sample

    workspace_id = uuid.UUID(fabric_info["workspace_id"])

    import_pbix_sample(["AdventureWorksSempy"], workspace_id, skip_report=False)

    return fabric_info



from ._fabric_data_agent_api import FabricDataAgentAPI
from ._tagged_value import TaggedValue
from sempy.fabric.exceptions import FabricHTTPException

from uuid import uuid4

import logging
import pandas as pd
import time
import typing as t
import warnings


# The list of selectable element types (e.g. columns, measures are not selectable since the UX doesn't support it)
SELECTABLE_ELEMENT_TYPES = [
    "semantic_model.table",
    "lakehouse_tables.table",
    "warehouse_tables.table",
    "kusto.table",
]

# The list of datasource types that allow configuration updates
CONFIGURABLE_DATASOURCE_TYPES = [
    "lakehouse_tables",
    "data_warehouse",
    "kusto",
]

# The list of datasource types that allow table/column description
DESCRIPTION_ENABLED_DATASOURCE_TYPES = [
    "lakehouse_tables",
    "data_warehouse",
]

# The list of element types that allow table/column description
DESCRIPTION_ENABLED_ELEMENT_TYPES = [
    "lakehouse_tables.table",
    "warehouse_tables.table",
    "lakehouse_tables.column",
    "warehouse_tables.column",
]


class Datasource:
    """
    Represents a datasource within an DataAgent.

    Attributes
    ----------
    _client : FabricDataAgentAPI
        The FabricDataAgentAPI client instance.
    _id : str
        The unique identifier of the datasource.
    """

    _client: FabricDataAgentAPI
    _id: str

    def __init__(self, client: FabricDataAgentAPI, id: str) -> None:
        """
        Initialize a Datasource instance.

        Parameters
        ----------
        client : FabricDataAgentAPI
            The FabricDataAgentAPI client to interact with the datasource.
        id : str
            The unique identifier of the datasource.
        """
        self._client = client
        self._id = id

    def __repr__(self):
        """
        Return a string representation of the Datasource.

        Returns
        -------
        str
            The string representation including the datasource ID.
        """
        return f"Datasource({self._id})"

    def pretty_print(self, include_type: t.Optional[bool] = False) -> None:
        """
        Pretty print the datasource configuration.

        Parameters
        ----------
        include_type : bool, optional
            Whether to include the type of the element in the output. Defaults to False.
        """
        config = self._client.get_datasource(self._id)

        def render_children(elements, level):
            if elements is None or len(elements) == 0:
                return

            indent = ' ' * 2

            for elem in elements:
                msg = f"{indent}|" * level + f' {elem["display_name"]}'

                if include_type:
                    msg += f' ({elem["type"]})'

                if elem.get("type") in SELECTABLE_ELEMENT_TYPES and elem["is_selected"]:
                    msg += " *"

                print(msg)
                render_children(elem["children"], level + 1)

        render_children(config.value.get("elements", []), 0)
        return config.value['display_name']

    def get_fewshots(self) -> pd.DataFrame:
        """
        Get the fewshots for the datasource.

        Returns
        -------
        pd.DataFrame
            A DataFrame containing the fewshots (Id, Question, Query, State, Embedding).
        """
        data = self._client.get_datasource_fewshots(self._id)

        return pd.DataFrame(
            [
                {
                    "Id": f["id"],
                    "Question": f["question"],
                    "Query": f["query"],
                    "State": f["state"],
                    "Embedding": f["embedding"],
                }
                for f in data.value.get("fewShots", [])
            ],
            columns=["Id", "Question", "Query", "State", "Embedding"],
        )

    def add_fewshots(self, fewshots: dict) -> None:
        """
        Add fewshots to the datasource.

        Parameters
        ----------
        fewshots : dict
            The dictionary of question and query.
        """
        datasource_fewshots = self._client.get_datasource_fewshots(self._id)

        data = datasource_fewshots.value.get("fewShots", [])

        for question, query in fewshots.items():
            data.append(
                {
                    "Id": str(uuid4()),
                    "Index": len(data),
                    "Question": question,
                    "Query": query,
                }
            )

        parentId = datasource_fewshots.value.get("parentId")
        if not parentId:
            datasource_fewshots.value["parentId"] = self._id

        datasource_fewshots.value["fewShots"] = data

        self._client.set_datasource_fewshots(self._id, datasource_fewshots)

    def remove_fewshot(self, fewshot_id: str) -> None:
        """
        Remove a fewshot from the datasource.

        Parameters
        ----------
        fewshot_id : str
            The ID of the fewshot to remove.
        """
        fewshots = self._client.get_datasource_fewshots(self._id)

        data = fewshots.value.get("fewShots", [])

        # remove the fewshot
        data = [f for f in data if f["id"] != fewshot_id]

        fewshots.value["fewShots"] = data

        self._client.set_datasource_fewshots(self._id, fewshots)

    def get_configuration(self) -> dict:
        """
        Get the configuration of the datasource.

        Returns
        -------
        dict
            The configuration of the datasource.
        """
        return self._client.get_datasource(self._id).value

    def update_configuration(
        self,
        instructions: str | None = None,
        schema_mode: str | None = None,
        user_description: str | None = None,
    ) -> None:
        """
        Update the configuration of the datasource.

        Parameters
        ----------
        instructions : str, optional
            Additional instructions for the datasource.
        schema_mode : str, optional
            The schema mode.
        user_description : str, optional
            The user description.
        """
        config = self._client.get_datasource(self._id)
        datasource_type = config.value.get("type")

        if datasource_type not in CONFIGURABLE_DATASOURCE_TYPES:
            raise ValueError(
                f"Datasource type '{datasource_type}' does not support configuration updates. Only {CONFIGURABLE_DATASOURCE_TYPES} are supported."
            )

        if instructions:
            config.value["additional_instructions"] = instructions

        if schema_mode:
            config.value["schema_mode"] = schema_mode

        if user_description:
            config.value["user_description"] = user_description

        self._client.set_datasource(config)

    def _set_is_selected(self, selected: bool, *path: t.Union[str, t.List[str]]) -> TaggedValue:
        """
        Set the selection status of a specific path in the datasource.

        Parameters
        ----------
        selected : bool
            The selection status to set.
        *path : str
            The path to set the selection status for.

        Returns
        -------
        TaggedValue
            The updated configuration.
        """
        config = self._client.get_datasource(self._id)

        # recursive function to set the selection status
        def recurse(l_path: t.List[str], elements: t.List[dict] | None):
            if l_path is None or len(l_path) == 0 or elements is None:
                return

            for elem in elements:
                if elem["display_name"] == l_path[0]:
                    if len(l_path) == 1:
                        elem_type = elem.get("type")
                        if elem_type not in SELECTABLE_ELEMENT_TYPES:
                            raise ValueError(f"Only table elements can be selected: '{elem_type}'")

                        elem["is_selected"] = selected
                    else:
                        recurse(l_path[1:], elem["children"])

                    return

            warnings.warn(f"Path {path} not found in datasource")

        elements = config.value.get("elements")

        if isinstance(path[0], list):
            for p in path:
                recurse(t.cast(t.List[str], p), elements)
        else:
            recurse(list(t.cast(t.Tuple[str, ...], path)), elements)

        return config

    def _set_is_selected_all(self, selected: bool) -> TaggedValue:
        """
        Set the selection status for all elements in the datasource.

        Parameters
        ----------
        selected : bool
            The selection status to set.

        Returns
        -------
        TaggedValue
            The updated configuration.
        """
        config = self._client.get_datasource(self._id)

        def set_selected(elements):
            for elem in elements:
                if elem.get("type") in SELECTABLE_ELEMENT_TYPES:
                    elem["is_selected"] = selected

                set_selected(elem["children"])

        set_selected(config.value["elements"])

        return config

    def _select(self, is_selected: bool, *path: t.Union[str, t.List[str]]) -> None:
        """
        Select or unselect elements in the datasource.

        Parameters
        ----------
        is_selected : bool
            Whether to select or unselect the elements.
        *path : str
            The path to select or unselect. e.g. self._select(True, "dbo", "table1", "col1").
        """
        if len(path) == 0:
            config = self._set_is_selected_all(is_selected)
        else:
            config = self._set_is_selected(is_selected, *path)

        self._client.set_datasource(config)

    def select(self, *path: t.Union[str, t.List[str]]) -> None:
        """
        Select elements in the datasource.

        Parameters
        ----------
        *path : str
            The path to select.
        """
        self._select(True, *path)

    def unselect(self, *path: t.Union[str, t.List[str]]) -> None:
        """
        Unselect elements in the datasource.

        Parameters
        ----------
        *path : str
            The path to unselect.
        """
        self._select(False, *path)

    def update_description(
        self,
        descriptions_by_path: dict[tuple[str, ...], str | None],
    ) -> None:
        """
        Update the description of the tables and columns in the datasource.

        Parameters
        ----------
        descriptions_by_path : dict[tuple[str, ...], str | None]
            A dictionary where the key is a tuple of strings representing the path to the element
            and the value is the description to set. e.g. self.update_description({("dbo", "table1", "col1"): "col1 description"}).
        """

        config = self._client.get_datasource(self._id)
        datasource_type = config.value.get("type")

        if datasource_type not in DESCRIPTION_ENABLED_DATASOURCE_TYPES:
            raise ValueError(
                f"Datasource type '{datasource_type}' does not support description updates. Only {DESCRIPTION_ENABLED_DATASOURCE_TYPES} are supported."
            )

        for path, description in descriptions_by_path.items():
            elements = config.value.get("elements", [])
            current = elements
            target_element = None
            for name in path:
                found = False
                for candidate in current:
                    if candidate["display_name"] == name:
                        target_element = candidate
                        current = candidate.get("children", [])
                        found = True
                        break
                if not found:
                    raise ValueError(f"Path {path} not found in datasource")

            if target_element is not None:
                if target_element.get("type") not in DESCRIPTION_ENABLED_ELEMENT_TYPES:
                    raise ValueError(
                        f"Only table and column elements can have descriptions: '{target_element.get('type')}'"
                    )
                target_element["description"] = description

        self._client.set_datasource(config)



from concurrent.futures import ThreadPoolExecutor
from datetime import datetime
from IPython.display import display, HTML, Markdown
from openai import NotFoundError
from pandas.io.formats.style import Styler
from tqdm import tqdm
from typing import Optional, Any

from fabric.dataagent.client import FabricDataAgentManagement
from fabric.dataagent.client import FabricOpenAI
from sempy.fabric.exceptions import FabricHTTPException
from fabric.dataagent.datasources.base import BaseSource
from fabric.dataagent.datasources import make_source
from fabric.dataagent.client._util import resolve_workspace_name_and_id
from synapse.ml.internal_utils.session_utils import get_fabric_context

import json
import logging
import numpy as np
import pandas as pd
import re
import time
import uuid
import string

OPEN_AI_MODEL = "gpt-4o"
USER_ROLE = "user"
DATA_AGENT_STAGE = "production"
SPARK_HOST = "spark.trident.pbiHost"

def evaluate_data_agent(
    df: pd.DataFrame,
    data_agent_name: str,
    workspace_name: Optional[str] = None,
    table_name: str = 'evaluation_output',
    critic_prompt: Optional[str] = None,
    data_agent_stage: str = DATA_AGENT_STAGE,
    max_workers: int = 5,
    num_query_repeats: int = 1
):
    """
    API to evaluate the Data Agent. Returns the unique id for the evaluation run.

    Parameters
    ----------
    df : pd.DataFrame
        DataFrame with question and expected_answer list.
    data_agent_name : str
        Name of the Data Agent.
    workspace_name : str, optional
        Workspace Name if Data Agent is in different workspace. Default to None.
    table_name : str, optional
        Table name to store the evaluation result. Default to 'evaluation_output'.
    critic_prompt : str, optional
        Prompt to evaluate the actual answer from Data Agent. Default to None.
    data_agent_stage : str, optional
        Data Agent stage ie., sandbox or production. Default to production.
    max_workers: int, optional
        Maximun worker nodes that need to run parallely. Default to 5.
    num_query_repeats: int, optional
        Number of times to evaluate each question. Default is 1.

    Returns
    -------
    str
        Unique id for the evaluation run.
    """

    output_rows = []
    output_steps = []
    run_timestamp = datetime.now().replace(tzinfo=None)
    # Unique id for each evaluation
    eval_id = str(uuid.uuid4())

    # Create Data Agent management
    data_agent = FabricDataAgentManagement(data_agent_name, workspace_name)

    from concurrent.futures import as_completed

    # Prepare arguments for each row
    rows = df.to_dict(orient="records")
    futures = []
    failed_urls = []
    display_handle = display(Markdown(""), display_id=True)

    with ThreadPoolExecutor(max_workers=max_workers) as executor:
        for row in rows:
            for _ in range(num_query_repeats):
                futures.append(
                    executor.submit(
                        evaluate_row,
                        row,
                        data_agent_name,
                        workspace_name,
                        data_agent_stage,
                        critic_prompt,
                        eval_id,
                        run_timestamp
                    )
                )
        for row in tqdm(as_completed(futures), total=len(futures)):
            output_row, output_step = row.result()
            output_rows.append(output_row)
            output_steps.append(output_step)
            if (
                "evaluation_judgement" in output_row
                and (
                    output_row["evaluation_judgement"] is False
                    or pd.isna(output_row["evaluation_judgement"])
                )
                and "thread_url" in output_row
                and output_row["thread_url"]
            ):
                failed_urls.append((output_row["thread_url"], output_row.get("thread_id", "")))
                bullet_list = "\n".join([
                    f"- [{thread_id}]({url})" if thread_id else f"- [{url}]({url})"
                    for url, thread_id in failed_urls
                ])
                if failed_urls:
                    display_handle.update(Markdown(f"**🔗Failed Thread(s):**\n{bullet_list}"))
                else:
                    display_handle.update(Markdown(""))

    # Sort output_rows so all variations for a question are together and in order
    df = pd.DataFrame(output_rows)
    df = df.sort_values(['question']).reset_index(drop=True)

    # Add configuration and data sources to the DataFrame
    df_data_agent = add_data_agent_details(df, data_agent)

    # Saving the evaluation output to a file
    save_output(df_data_agent, table_name)
    # Saving the evaluation output steps to a file
    save_output(pd.DataFrame(output_steps), f"{table_name}_steps")

    return eval_id

def evaluate_row(
    row: dict,
    data_agent_name: str,
    workspace_name: Optional[str],
    data_agent_stage: str,
    critic_prompt: Optional[str],
    eval_id: str,
    run_timestamp: datetime
):
    """
    Evaluate a single row of input DataFrame.
    
    Parameters
    ----------
    row : dict
        A single row from the input DataFrame containing question and expected_answer.
    data_agent_name : str
        Name of the Data Agent.
    workspace_name : str, optional
        Workspace Name if Data Agent is in different workspace. Default to None.
    data_agent_stage : str
        Data Agent stage ie., sandbox or production. Default to production.
    critic_prompt : str, optional
        Prompt to evaluate the actual answer from Data Agent. Default to None.
    eval_id : str
        Unique id for the evaluation run.
    run_timestamp : datetime
        Timestamp indicating when the evaluation run started.
    
    Returns
    -------
    tuple[object, object]
        Formatted response of the Data Agent and run steps.
    """

    # Map user-friendly stage names to internal values
    stage_map = {
        'draft': 'sandbox',
        'sandbox': 'sandbox',
        'publishing': 'production',
        'production': 'production'
    }
    # Normalize input and get the internal stage value
    data_agent_stage_internal = stage_map.get(str(data_agent_stage).lower(), 'production')
    
    fabric_client = FabricOpenAI(
        artifact_name=data_agent_name,
        workspace_name=workspace_name,
        ai_skill_stage=data_agent_stage_internal
    )
    data_agent = FabricDataAgentManagement(data_agent_name, workspace_name)
    query: str = str(row['question'])
    expected_answer: str = str(row['expected_answer'])

    # Generate the response for the query
    output_row, run_steps = generate_answer(
        query, 
        fabric_client, 
        data_agent, 
        expected_answer, 
        critic_prompt, 
        eval_id, 
        run_timestamp
    )

    return output_row, run_steps

def generate_answer(
    query: str,
    fabric_client: FabricOpenAI, 
    data_agent: FabricDataAgentManagement, 
    expected_answer: str, 
    critic_prompt: Optional[str],
    eval_id: str,
    run_timestamp: datetime
):
    """
    Generates the response for input query.

    Parameters
    ----------
    query : str
        Question from the input DataFrame.
    fabric_client: FabricOpenAI
        An instance of the fabric client to interact with Data Agent.
    data_agent: FabricDataAgentManagement
        An instance of FabricDataAgentManagement to get the details of the Data Agent.
    expected_answer : str
        Expected answer from the input DataFrame.
    critic_prompt : str, optional
        Prompt to evaluate the actual answer from Data Agent. Default to None.
    eval_id : str
        Unique id for the evaluation run.
    run_timestamp : datetime
        Timestamp indicating when the evaluation run started.

    Returns
    -------
    tuple[object, object]
        Formatted response of the Data Agent and run steps.
    """

    # Unique id for each row in input dataset
    unique_id = str(uuid.uuid4())
    start_time = time.time()

    # Create thread with custom tag (uuid)
    thread = fabric_client.get_or_create_thread(unique_id)
    thread_id = thread.id

    # Generate answer for the input query
    message, run = get_message(fabric_client, thread_id, query)

    # Construct the thread URL
    thread_url = get_thread_url(thread_id, data_agent)

    run_steps = get_steps(fabric_client, thread_id, run.id, unique_id)

    # Generate the prompt for evaluating the actual answer
    prompt = generate_prompt(query, expected_answer, critic_prompt)

    # Generate answer for the evaluation prompt
    eval_message, eval_run = get_message(fabric_client, thread_id, prompt)
    eval_message_low = eval_message.lower()
    score = False if "no" in eval_message_low else True if "yes" in eval_message_low else np.nan
    
    end_time = time.time()
 
    output_row = {
        'id': unique_id,
        'evaluation_id': eval_id,
        'thread_id': thread_id,
        'run_timestamp': run_timestamp,
        'question': query,
        'expected_answer': expected_answer,
        'actual_answer': message,
        'execution_time_sec': round(end_time - start_time, 2),
        'status': run.status,
        'thread_url': thread_url,
        'evaluation_judgement': score,
        'evaluation_message': eval_message,
        'evaluation_status': eval_run.status,
        'evaluation_thread_url': thread_url
    }

    return output_row, run_steps


def get_message(fabric_client: FabricOpenAI, thread_id: str, query: str):
    """
    Get message for the input query and thread.

    Parameters
    ----------
    fabric_client: FabricOpenAI
        An instance of the fabric client created to interact with Data Agent.
    thread_id: str
        An unique identifier of the thread.
    query : str
        Question from the input DataFrame.

    Returns
    -------
    tuple[str, object]
        Tuple with actual answer and run instance.

    Raises
    -------
        RuntimeError: If fabric client is None.
    """

    # Raise RuntimeError if fabric client is None
    if fabric_client is None:
        logging.debug("Fabric client is None")
        raise RuntimeError("Fabric client is None")

    # Create assistant
    assistant = fabric_client.beta.assistants.create(model=OPEN_AI_MODEL)

    # Add message to the thread
    fabric_client.beta.threads.messages.create(
        thread_id=thread_id, role=USER_ROLE, content=query
    )

    try:
        # Start the run
        run = fabric_client.beta.threads.runs.create_and_poll(
            thread_id=thread_id, assistant_id=assistant.id
        )
    except NotFoundError:
        raise ValueError("Invalid input for data_agent_stage. Please use sandbox/draft if DataAgent is not published.")

    status = run.status
    # Log error message if status is failed
    if status == "failed" and run.last_error:
        generated_answer = f"Error ({run.last_error.code}): {run.last_error.message}"
        logging.debug(generated_answer)
    else:
        # Get the messages from response
        messages = fabric_client.beta.threads.messages.list(
            thread_id=thread_id, order="desc"
        )

        if messages:
            generated_answer = list(messages)[0].content[0].text.value
        else:
            generated_answer = "No answer returned from Data Agent"

    return generated_answer, run


def get_steps(
    fabric_client: FabricOpenAI, thread_id: str, run_id: str, unique_id: Optional[str] = None
):
    """
    Get steps for the run.

    Parameters
    ----------
    fabric_client : FabricOpenAI
        An instance of the fabric client created to interact with Data Agent.
    thread_id : str
        An unique identifier of the thread.
    run_id : str
        An unique identifier of the run.
    unique_id : str
        An unique identifier for the input processing row.

    Returns
    -------
    dict
        Dictionary of run steps in the Data Agent response.

    Raises
    -------
        RuntimeError: If fabric client is None.
    """

    # Raise RuntimeError if fabric client is None
    if fabric_client is None:
        logging.debug("Fabric client is None")
        raise RuntimeError("Fabric client is None")

    function_names = []
    function_queries = []
    function_outputs = []
    sql_commands = []
    dax_commands = []
    kql_commands = []

    # Get run steps for a thread
    run_steps = fabric_client.beta.threads.runs.steps.list(
        thread_id=thread_id, run_id=run_id
    )

    # Extract list of run steps
    for run_step in run_steps:
        if run_step.step_details.type == "tool_calls":
            for tool_call in run_step.step_details.tool_calls:
                if tool_call.type == "function":
                    function_names.append(str(tool_call.function.name))
                    arguments = json.loads(tool_call.function.arguments)
                    # Convert to json dict if arguments is not dict
                    if not isinstance(arguments, dict):
                        arguments = json.loads(arguments)
                    # Check if arguments is dict to get the params or save arguments as function query
                    if isinstance(arguments, dict):
                        query = arguments.get("query") or arguments.get("natural_language_query")
                    else:
                        query = arguments
                    function_queries.append(str(query))
                    function_output = tool_call.function.output
                    function_outputs.append(str(function_output))
                    commands = get_commands(function_output)
                    sql_commands.append(str(commands.get("sql")))
                    dax_commands.append(str(commands.get("dax")))
                    kql_commands.append(str(commands.get("kql")))

    return {
        'id': unique_id,
        'thread_id': thread_id,
        'run_id': run_id,
        'function_names': str(function_names),
        'function_queries': str(function_queries),
        'function_outputs': str(function_outputs),
        # TODO: Make seperate tables for the commands.
        'sql_steps': str(sql_commands),
        'dax_steps': str(dax_commands),
        'kql_steps': str(kql_commands),
    }


def get_commands(output: str):
    """
    Get commands from run steps.

    Parameters
    ----------
    output : str
        Output string from a Data Agent response.

    Returns
    -------
    dict
        Dictionary of command type and command value returned in run steps.
    """
    commands = {}
    # Regular expression pattern to extract content inside triple backticks
    pattern = r"```(sql|dax|kql)\s(.*?)```"

    if output:
        # Extract matches
        matches = re.findall(pattern, output, re.DOTALL)
        # Store extracted commands in a dictionary
        commands = {match[0]: match[1] for match in matches}

    return commands


def generate_prompt(
    query: str, expected_answer: str, critic_prompt: Optional[str] = None
):
    """
    Generate the prompt for the evaluation.

    Parameters
    ----------
    query : str
        Question from an input DataFrame.
    expected_answer : str
        Expected answer from an input DataFrame.
    critic_prompt : str, optional
        Prompt to evaluate the actual answer from Data Agent. Default to None.

    Returns
    -------
    str
        String prompt for the evaluation.
    """

    import textwrap

    if critic_prompt:
        prompt = critic_prompt.format(
            query=query, expected_answer=expected_answer
        )
    else:
        prompt = f"""
        Given the following query and ground truth, please determine if the most recent answer is equivalent or satifies the ground truth. If they are numerically and semantically equivalent or satify (even with reasonable rounding), respond with "Yes". If they clearly differ, respond with "No". If it is ambiguous or unclear, respond with "Unclear". Return only one word: Yes, No, or Unclear..

        Query: {query}

        Ground Truth: {expected_answer}
        """

    return textwrap.dedent(prompt)


def add_data_agent_details(df: pd.DataFrame, data_agent: FabricDataAgentManagement):
    """
    Add Data Agent details to the DataFrame.

    Parameters
    ----------
    df : pd.DataFrame
        DataFrame with processed output rows.
    data_agent: FabricDataAgentManagement
        An instance of FabricDataAgentManagement to get the details of the Data Agent.

    Returns
    -------
    pd.DataFrame
        Updated dataframe with Data Agent details.
    """

    # Return unmodified DataFrame if data agent is None
    if data_agent is None:
        return df

    df['data_agent_version'] = pd.Series(None, dtype='string')
    df['data_agent_etag'] = pd.Series(None, dtype='string')
    df['data_agent_last_updated'] = pd.Series(None, dtype='string')

    # Add the Data Agent details to DataFrame
    df['data_agent_configuration'] = str(data_agent.get_configuration())
    try:
        # Fetch the publish info of DataAgent
        publishing_info = data_agent._client.get_publishing_info()
        df['data_agent_version'] = publishing_info.value['currentVersion']
        df['data_agent_etag'] = publishing_info.etag
        df['data_agent_last_updated'] = publishing_info.value['lastUpdated']
    except FabricHTTPException:
        # Skipping the publish info as DataAgent is not published
        pass
    data_sources = str(data_agent.get_datasources())
    df['data_sources'] = [data_sources] * len(df)

    return df


def get_fabric_host():
    """
    Get Fabric host address.

    Returns
    -------
    str
        Fabric host address.

    Raises
    -------
        RuntimeError: If host address is None.
    """
    host = get_fabric_context().get(SPARK_HOST)
    if host is None:
        logging.debug(f"Fabric Host address is empty")
        raise RuntimeError("Fabric Host address is empty")

    return host.replace('api', "", 1)


def get_thread_url(thread_id: str, data_agent: FabricDataAgentManagement):
    """
    Get thread URL for the thread.

    Parameters
    ----------
    thread_id : str
        An unique identifier of the thread.
    data_agent : FabricDataAgentManagement
        An instance of FabricDataAgentManagement to get the details of the Data Agent.

    Returns
    -------
    str
        Thread URL for the input thread id.
    """

    data_agent_id = data_agent._client.data_agent_id
    host = get_fabric_host()

    thread_url = f"https://{host}/workloads/de-ds/dataagents/{data_agent_id}/externalThread?debug.aiSkillThreadIdOverride={thread_id}&debug.aiSkillViewPublishedOverride=0"

    return thread_url


def save_output(df: pd.DataFrame, table_name: str):
    """
    Saves the Dataframe to the Delta table.

    Parameters
    ----------
    df : pd.DataFrame
        DataFrame with processed output rows.
    table_name : str
        Table name to store the evaluation result
    """
    lakehouse_path = _default_lakehouse_path()
    table_path = f"{lakehouse_path}/Tables/{table_name}"

    if _on_jupyter():
        from deltalake.writer import write_deltalake

        write_deltalake(table_path, df, mode="append")
    else:
        from pyspark.sql import SparkSession
        from pyspark.sql.functions import col

        spark = SparkSession.builder.getOrCreate()
        try:
            # Get schema if table exists
            delta_df = spark.read.format("delta").load(table_path)
            expected_schema = delta_df.schema
            spark_df = spark.createDataFrame(df, schema=expected_schema)
        except Exception:
            spark_df = spark.createDataFrame(df)
            if "run_timestamp" in spark_df.columns:
                # Making the timestamp compatible with Python
                spark_df = spark_df.withColumn("run_timestamp", col("run_timestamp").cast("timestamp_ntz"))

        spark_df.write.format("delta").mode("append").save(table_path)

def get_evaluation_summary(
    table_name: str = 'evaluation_output', verbose: bool = False
):
    """
    Overall summary of an evaluation stored in the delta table.

    Parameters
    ----------
    table_name : str, optional
        Table name to store the evaluation result. Default to 'evaluation_output'.
    verbose : bool, optional
        Flag to print the evaluation summary. Default to False.

    Returns
    -------
    pd.DataFrame
        DataFrame with summary details.
    """

    # Get the delta table
    df = _get_data(table_name)

    # Return if table is not found
    if df is None:
        return None

    # Calculate the percentage of True values in the DataFrame
    eval_percentage = (df['evaluation_judgement'].mean()) * 100

    # Group by timestamp and count the occurrences of True, False, and None in the 'evaluation_judgement' column
    grouped_df = (
        df.groupby('evaluation_id')['evaluation_judgement']
        .value_counts(dropna=False)
        .unstack(fill_value=0)
    )

    # Reindex to ensure True, False, None columns are included even if their count is zero
    grouped_df = grouped_df.reindex(columns=[True, False, np.nan], fill_value=0)

    # Calculate the percentage of True values per evaluation
    grouped_df['%'] = ((grouped_df[True] / grouped_df.sum(axis=1)) * 100).round(1)

    # Reset index and rename the columns in the DataFrame
    grouped_df = grouped_df.reset_index().rename(
        columns={True: "T", False: "F", np.nan: "?"}
    )

    grouped_df.columns.name = 'index'

    if verbose:
        eval_string = f"<h5>Evaluation judgement in percentage: {int(eval_percentage) if eval_percentage.is_integer() else round(eval_percentage, 1)}%</h5>"
        display(HTML(eval_string))
        styled_df = grouped_df.style.format(
            {
                '%': '{:.1f}'
            },
            escape="html"
        )
        _display_styled_html(styled_df)

    return grouped_df


def get_evaluation_details(
    evaluation_id: str,
    table_name: str = 'evaluation_output',
    get_all_rows: bool = False,
    verbose: bool = False,
):
    """
    Get evaluation details of a single run.

    Parameters
    ----------
    evaluation_id : str
        Unique id for the evaluation run.
    table_name : str, optional
        Table name to store the evaluation result. Default to 'evaluation_output'.
    get_all_rows : bool, optional
        Flag to get all the rows for an evaluation. Default to False, which returns only failed evaluation rows.
    verbose : bool, optional
        Flag to print the evaluation summary. Default to False.

    Returns
    -------
    pd.DataFrame
        DataFrame with single evaluation details.
    """

    # Get the delta table
    df = _get_data(table_name)

    # Return if table is not found
    if df is None:
        return None

    df = df.sort_values(['question']).reset_index(drop=True)

    filtered_df = df[df["evaluation_id"] == evaluation_id]
    # Filter for only failed rows if get_all_rows is False
    if not get_all_rows:
        filtered_df = filtered_df[(filtered_df["evaluation_judgement"] == False) | (pd.isna(filtered_df["evaluation_judgement"]))]

    filtered_df = filtered_df.reset_index(drop=True)
    filtered_df.columns.name = 'index'
    if verbose:
        if filtered_df.empty:
            display(HTML('<b>There are no failed evaluation rows.</b> Use <i>get_all_rows</i> parameter as <i>True</i> to display all the evaluation rows.'))
            return filtered_df
        
        selected_cols_df = filtered_df[
            [
                "question",
                "expected_answer",
                "actual_answer",
                "evaluation_judgement",
                "thread_id",
                "thread_url"
            ]
        ]

        thread_urls = []
        for index, row in selected_cols_df.iterrows():
            thread_urls.append(
                f'[{row["thread_id"]}]({row["thread_url"]})'
            )

        # Make a copy once
        selected_cols_df = selected_cols_df.copy()

        # Assign columns in a vectorized way
        selected_cols_df['thread_url'] = thread_urls
        selected_cols_df['evaluation_judgement'] = selected_cols_df['evaluation_judgement'].fillna("Unclear")
        selected_cols_df = selected_cols_df.drop(columns=['thread_id'])
        styled_cols_df = selected_cols_df.style.format(
            {
                "question": _markdown_formatter,
                "expected_answer": _markdown_formatter,
                "actual_answer": _markdown_formatter,
                "thread_url": _markdown_formatter
            },
            escape="html"
        )

        _display_styled_html(styled_cols_df)

    return filtered_df

def get_evaluation_summary_per_question(
    evaluation_id: Optional[str] = None,
    table_name: str = 'evaluation_output',
    verbose: bool = False
):
    """
    Summary of evaluation results per question for a specific evaluation_id, showing the percentage of True for each question based on all its variations.

    Parameters
    ----------
    evaluation_id : str, optional
        Unique id for the evaluation run. If None, it returns the summary for all evaluations.
    table_name : str, optional
        Table name to store the evaluation result. Default to 'evaluation_output'.
    verbose : bool, optional
        Flag to print the evaluation summary. Default to False.

    Returns
    -------
    pd.DataFrame
        DataFrame with summary details per question for the given evaluation_id.
    """
    df = _get_data(table_name)
    if df is None or 'question' not in df.columns:
        return None
    
    if evaluation_id is not None:
        df = df[df['evaluation_id'] == evaluation_id]
    
    if df.empty:
        return None
    
    # Group by question and count the occurrences of True, False, and None in the 'evaluation_judgement' column
    # and calculate the percentage of True values per question
    grouped_df = (
        df.groupby('question')['evaluation_judgement']
        .value_counts(dropna=False)
        .unstack(fill_value=0)
    )
    grouped_df = grouped_df.reindex(columns=[True, False, np.nan], fill_value=0)
    grouped_df['%'] = ((grouped_df[True] / grouped_df.sum(axis=1)) * 100).round(1)
    grouped_df = grouped_df.reset_index().rename(
        columns={True: "T", False: "F", np.nan: "?"}
    )

    # Extract failed records for each question
    # and format the thread URLs
    failed_records = (
        df[(df["evaluation_judgement"] == False) | (pd.isna(df["evaluation_judgement"]))]
        .groupby("question")
        .apply(lambda x: x.to_dict(orient="records"))
        .reset_index()
        .rename(columns={0: "failed_records"})
    )

    failed_records[["failed_threads", "failed_thread_urls"]] = failed_records["failed_records"].apply(
        _extract_failed_thread_info
    ).apply(pd.Series)

    merged_df = grouped_df.merge(failed_records, on="question", how="left")
    merged_df['failed_threads'] = merged_df['failed_threads'].fillna("No failed records")
    merged_df.columns.name = 'index'

    if verbose:
        display(HTML(f"<h5>Evaluation summary per question:</h5>"))
        display_cols = ['T', 'F', '?', '%', 'failed_threads', 'question']
        styled_df = merged_df[display_cols].style.format(
            {
                'question': _markdown_formatter,
                'failed_threads': _markdown_formatter,
                '%': '{:.1f}'
            },
            escape="html"
        )
        _display_styled_html(styled_df)

    df_cols = ['T', 'F', '?', '%', 'failed_thread_urls', 'question']
    df = merged_df[df_cols]
    return df

def _extract_failed_thread_info(records):
    """
    Extracts and formats thread URLs from a list of record dictionaries.
    Iterates over the provided records, checks for the presence of a non-null "thread_url" key,
    and constructs an HTML string of numbered links as well as a list of the URLs.
    
    Parameters
    ----------
    records: list of dict
        A list of dictionaries, each potentially containing a "thread_url" key.

    Returns
    -------
    tuple:
        hrefs: str
            An HTML string with numbered anchor tags linking to each thread URL.
        urls: list of str
            A list of thread URLs extracted from the records.
    """
    hrefs = ""
    urls = []
    for i, rec in enumerate(records):
        if pd.notna(rec.get("thread_url")):
            hrefs += f'[{i+1}]({rec["thread_url"]}) '
            urls.append(rec["thread_url"])

    return hrefs, urls

def _default_lakehouse_path():
    """
    Get default lakehouse path.

    Returns
    -------
    str
        Default lakehouse path.
    """
    defaultFs = get_fabric_context().get('fs.defaultFS')
    lakehouse_id = get_fabric_context().get('trident.lakehouse.id')
    return f"{defaultFs}{lakehouse_id}"


def _get_data(table_name: str):
    """
    Get data from the specified delta table and return it as a Pandas DataFrame.

    Parameters
    ----------
    table_name : str
        Table name which contains the evaluation result.

    Returns
    -------
    DataFrame
        Data from the specified delta table as a Pandas DataFrame instance.
    """
    import os

    lakehouse_path = _default_lakehouse_path()
    table_path = f"{lakehouse_path}/Tables/{table_name}"
    df = None
    try:
        if _on_jupyter():
            from deltalake import DeltaTable

            # Load the Delta Lake table
            delta_table = DeltaTable(table_path)
            df = delta_table.to_pandas() if delta_table is not None else None
        else:
            from pyspark.sql import SparkSession

            spark = SparkSession.builder.getOrCreate()
            spark_df = spark.read.format("delta").load(table_path)
            df = spark_df.toPandas()
    except Exception:
        message = f"<h4>Table does not exist. Please provide the table name from attached default lakehouse.</h4>"
        display(HTML(message))
        return None

    return df

def _on_jupyter() -> bool:
    import os
    return os.environ.get("MSNOTEBOOKUTILS_RUNTIME_TYPE", "").lower() == "jupyter"

def _markdown_formatter(text):
    """ Convert text to HTML using markdown2.
    Parameters 
    ----------
    text : str
        The text to convert.
    """
    import markdown2

    if pd.isna(text):
        return ""
    return markdown2.markdown(text)

def _sql_formatter(text):
    """ Convert SQL text to HTML using markdown2 with fenced code blocks.
    Parameters
    ----------
    text : str
        The SQL text to convert.
    """
    import markdown2

    if pd.isna(text):
        return ""
    return markdown2.markdown(f"```sql\n{text}\n```", extras=["fenced-code-blocks"])

def _display_styled_html(styler: Styler):
    """
    Display styled html of a Dataframe.

    Parameters
    ----------
    styler : Styler
        A pandas Styler object to style.
    """

    # Convert DataFrame to HTML and add inline CSS to enforce left alignment
    styled_html = styler \
        .set_properties(**{'text-align': 'left'}) \
        .set_table_styles(
            [{'selector': 'th', 'props': [('text-align', 'left')]}]
        )

    display(styled_html)


def _get_source(
    datasource_id_or_name: str,  # name **or** id
    data_agent: FabricDataAgentManagement,
    workspace_id_or_name: str | None = None,  # optional
) -> BaseSource:
    """
    Return the datasource's type or raise if not found.
    Parameters
    ----------
    datasource_id_or_name : str
        The name or ID of the datasource.
    workspace_id_or_name : str, optional
        The workspace name or ID. Defaults to None. If `workspace_id_or_name` is supplied the search is performed in that
    workspace; otherwise the current workspace is used.
    data_agent : FabricDataAgentManagement
        An instance of FabricDataAgentManagement to get the details of the Data Agent.
    Returns
    -------
    Return a concrete datasource wrapper (BaseSource subclass), looking in the supplied workspace when given.

    Raises
    -------
    ValueError
        If the data source is not found.
    """
    # Normalize workspace
    ws_name, _ = resolve_workspace_name_and_id(workspace_id_or_name)

    # 2. fetch all datasources visible to the agent (may span workspaces)
    # pick the matching datasource from the agent list
    for ds in data_agent.get_datasources():
        cfg = ds.get_configuration()
        if cfg.get("id") == datasource_id_or_name or cfg.get("display_name") == datasource_id_or_name:
            # inject workspace name so connectors can use it
            cfg["workspace_name"] = ws_name
            return make_source(cfg)

    raise ValueError(
        f"Datasource '{datasource_id_or_name}' (workspace={workspace_id_or_name}) not found"
    )


def _extract_placeholders(template: str) -> set[str]:
    """
    Return all *named* placeholders in a str.format template.
    """
    fmt = string.Formatter()
    # correct tuple order: literal_text, field_name, format_spec, conversion
    return {
        field_name
        for _lit, field_name, _spec, _conv in fmt.parse(template)
        if field_name
    }


def add_ground_truth(
    question: str,
    answer_template: str,
    datasource_id_or_name: str,
    query: str,
    data_agent: FabricDataAgentManagement,
    exec_ctx: Optional[Any] = None,
    *,
    source: Optional[BaseSource] = None,
    verbose: bool = False
) -> pd.DataFrame:
    """
    Update ground-truth for one query / answer pair.
    Works with Lakehouse, Warehouse and Kusto datasources.

    The SQL query must return exactly one row.
    The answer template must use **named** placeholders that match the
    column names of that row, e.g.  "Total {sales} in {country}".

    Parameters
    ----------
    question : str
        Question from the input DataFrame.
    answer_template : str
        Template for the expected answer.
    datasource_id_or_name : str
        Name of the data source.
    query : str
        Query to get the expected answer.
    data_agent : FabricDataAgentManagement
        An instance of FabricDataAgentManagement to get the details of the Data Agent.
    exec_ctx : Any, optional
        Execution context to reuse an open connection. Default is None, which creates a new connection.
    source : BaseSource, optional
        A datasource object to reuse. If not provided, it will be fetched using `datasource_id_or_name` and `data_agent`.
    verbose : bool, optional
        Flag to print verbose output. Default is False.

    Returns
    -------
    pd.DataFrame
        DataFrame with updated ground truth.
    """
    # get or reuse the datasource
    if source is None:
        source = _get_source(datasource_id_or_name, data_agent)


    # check connection and execute query
    if exec_ctx is None:
        with source.connect() as ctx:
            df_res = ctx.query(query)
    else:
        df_res = exec_ctx.query(query)

    # check if the query returned exactly one row
    if df_res.empty:
        raise ValueError("Query returned no rows; cannot fill template")

    if len(df_res) > 1:
        raise ValueError(
            "Query returned multiple rows. " "add_ground_truth expects exactly one row."
        )

    if "{}" in answer_template:
        raise ValueError("Positional '{}' is not supported; use named placeholders")

    placeholders = _extract_placeholders(answer_template)

    missing = placeholders.difference(df_res.columns)
    if missing:
        raise ValueError(f"Missing columns in answer template: {', '.join(missing)}")
    
    # check if the answer template can be formatted with the row
    try:
        rendered = answer_template.format(**df_res.iloc[0].to_dict())
    except (KeyError, ValueError) as exc:
        raise ValueError(f"Template formatting failed: {exc}") from exc


    ground_truth_df = pd.DataFrame(
        {
            "question": [question],
            "expected_answer": [rendered],
            "datasource_id_or_name": [datasource_id_or_name],
            "query": [query],
        }
    )

    if verbose:
        styled_df = ground_truth_df.style.format(
            {
                "question": _markdown_formatter,
                "expected_answer": _markdown_formatter,
                "query": _sql_formatter
            },
            escape="html"
        )
        _display_styled_html(styled_df)

    return ground_truth_df


def add_ground_truth_batch(
    df: pd.DataFrame,  # cols: question, answer_template, query
    datasource_id_or_name: str,
    data_agent: FabricDataAgentManagement,
    verbose: bool = False,
) -> pd.DataFrame:
    """ Add ground truth for a batch of queries and answer templates.
    
    Parameters
    ----------
    df : pd.DataFrame
        DataFrame with columns: question, answer_template, query.
    datasource_id_or_name : str
        Name or ID of the data source.
    data_agent : FabricDataAgentManagement
        An instance of FabricDataAgentManagement to get the details of the Data Agent.
    verbose : bool, optional
        Flag to print verbose output. Default is False.

    Returns
    -------
    pd.DataFrame
        DataFrame with the ground truth added.
    """
    source = _get_source(datasource_id_or_name, data_agent)
    out = []
    with source.connect() as ctx:
        for r in tqdm(
            df.itertuples(index=False), total=df.shape[0], desc="ground-truth batch"
        ):
            out.append(                
                add_ground_truth(
                    str(r.question),
                    str(r.answer_template),
                    datasource_id_or_name,
                    str(r.query),
                    data_agent,
                    exec_ctx=ctx,  # reuse open connection
                    source=source,     # reuse datasource object
                )
            )
    ground_truth_df = pd.concat(out, ignore_index=True)
    ground_truth_df.columns.name = 'index'

    if verbose:
        styled_df = ground_truth_df.style.format(
            {
                "question": _markdown_formatter,
                "expected_answer": _markdown_formatter,
                "query": _sql_formatter
            },
            escape="html"
        )
        _display_styled_html(styled_df)

    return ground_truth_df



"""
Unit-tests for evaluator_api.update_ground_truth

They avoid heavy dependencies (deltalake, sempy_labs, Spark, HTTP) by
registering minimal stub modules in ``sys.modules`` **before** the code
under test is imported.
"""

from __future__ import annotations

import sys
import types
import pandas as pd
import pytest
import string
import importlib

def _ensure(name: str) -> types.ModuleType:
    """
    Return the real module if importable, otherwise create a dummy one.
    """
    if name in sys.modules:
        return sys.modules[name]
    try:
        return importlib.import_module(name)          # real package present
    except ModuleNotFoundError:
        mod = types.ModuleType(name)
        sys.modules[name] = mod
        return mod

# ── only the packages required by kustodb.py ──────────────────────────────
for path in (
    "azure.kusto",
    "azure.kusto.data",
    "azure.kusto.data.helpers",
    "azure.kusto.data.exceptions",
    "synapse.ml.fabric.token_utils",
    "synapse.ml.fabric.service_discovery",
):
    parts = path.split(".")
    for i in range(1, len(parts) + 1):
        _ensure(".".join(parts[:i]))

# minimal attributes for kustodb
az_data = _ensure("azure.kusto.data")
exc_mod = sys.modules["azure.kusto.data.exceptions"]
exc_mod.KustoServiceError = type("KustoServiceError", (), {})
exc_mod.KustoClientError  = type("KustoClientError",  (), {})
az_data.ClientRequestProperties = type(
    "ClientRequestProperties", (), {"set_option": lambda *_, **__: None}
)
az_data.KustoConnectionStringBuilder = type(
    "KustoConnectionStringBuilder",
    (),
    {"with_token_provider": staticmethod(lambda *_, **__: None)},
)
az_data.KustoClient = type(
    "KustoClient",
    (),
    {"execute": lambda *_, **__: types.SimpleNamespace(primary_results=[pd.DataFrame()])},
)
sys.modules["azure.kusto.data.helpers"].dataframe_from_result_table = (
    lambda *_: pd.DataFrame()
)

_ensure("synapse.ml.fabric.token_utils").TokenUtils = type(
    "TokenUtils",
    (),
    {
        "get_access_token": lambda *_, **__: "token",
        "get_aad_token": lambda *_, **__: "aad-token",
    },
)

_ensure("synapse.ml.fabric.service_discovery").get_fabric_env_config = (
    lambda *_a, **_kw: {}
)       


# 1.  Stub external libraries so evaluator_api imports cleanly
def _register_stub(name: str) -> types.ModuleType:
    mod = types.ModuleType(name)
    sys.modules[name] = mod
    return mod


# create parent & sub-modules (deltalake, deltalake.writer, sempy_labs)
for missing in ("deltalake", "deltalake.writer", "sempy_labs"):
    parts = missing.split(".")
    for i in range(1, len(parts) + 1):
        _register_stub(".".join(parts[:i]))

# give the required attribute so `from deltalake.writer import write_deltalake` works
sys.modules["deltalake.writer"].write_deltalake = lambda *_, **__: None

# 2.  Import module under test
from fabric.dataagent.evaluation import evaluator_api


# 3.  Helpers
def _dummy_sql(df: pd.DataFrame):
    """Mimic labs.ConnectX context manager with a .query() method returning *df*."""

    class _Ctx:
        def __enter__(self):
            return self

        def __exit__(self, exc_type, exc, tb):
            pass

        def query(self, _):
            return df

    return _Ctx()


# Patch the evaluator_api module to use our dummy SQL connection
def _patch(monkeypatch, df: pd.DataFrame):
    # force a lakehouse source and dummy connection
    monkeypatch.setattr(
        evaluator_api,
        "_get_source",
        lambda *_: types.SimpleNamespace(connect=lambda: _dummy_sql(df)),
    )


# 4.  Tests
def test_single_scalar_named_placeholder(monkeypatch):
    """Single column, named placeholder is rendered correctly."""
    _patch(monkeypatch, pd.DataFrame({"total": [42]}))

    out = evaluator_api.add_ground_truth(
        question="Total sales?",
        answer_template="Total sales: {total}",
        datasource_id_or_name="Any",
        query="dummy",
        data_agent=None,
    )
    assert out["expected_answer"].iloc[0] == "Total sales: 42"


def test_single_row_multi_column(monkeypatch):
    """Several columns with matching named placeholders."""
    _patch(monkeypatch, pd.DataFrame({"country": ["US"], "sales": [123]}))

    out = evaluator_api.add_ground_truth(
        "By country",
        "{country} sold {sales}",
        "Any",
        "dummy",
        data_agent=None,
    )
    assert out["expected_answer"].iloc[0] == "US sold 123"


def test_error_positional_placeholder(monkeypatch):
    """Positional '{}' is now forbidden."""
    _patch(monkeypatch, pd.DataFrame({"x": [1]}))

    with pytest.raises(ValueError, match="Positional"):
        evaluator_api.add_ground_truth(
            "Bad",
            "Answer is {}",
            "Any",
            "dummy",
            data_agent=None,
        )


def test_error_multiple_rows(monkeypatch):
    """Multiple query rows should raise."""
    _patch(monkeypatch, pd.DataFrame({"id": [1, 2]}))

    with pytest.raises(ValueError, match="multiple rows"):
        evaluator_api.add_ground_truth(
            "Bad",
            "{id}",
            "Any",
            "dummy",
            data_agent=None,
        )

# 5.  _extract_placeholders() edge-cases coverage
@pytest.mark.parametrize(
    "template, expected",
    [
        ("Value is {amount:.2f}", {"amount"}),          # format-spec
        ("Balance {{USD}} {balance}", {"balance"}),     # escaped braces
        ("Repeated {a}-{a}-{b}", {"a", "b"}),           # duplicates collapse
        ("No placeholders here", set()),                # no placeholders
    ],
)
def test_extract_placeholders_edge_cases(template, expected):
    assert evaluator_api._extract_placeholders(template) == expected

def test_add_ground_truth_batch(monkeypatch):
    """add_ground_truth_batch returns one output row per input row."""
    df_single = pd.DataFrame({"val": [1]})            # 1-row result for every query
    _patch(monkeypatch, df_single)

    batch_in = pd.DataFrame(
        [
            dict(question="q1", answer_template="v={val}", query="dummy"),
            dict(question="q2", answer_template="v={val}", query="dummy"),
        ]
    )

    out = evaluator_api.add_ground_truth_batch(batch_in, "Any", data_agent=None)

    assert len(out) == 2
    assert list(out["question"]) == ["q1", "q2"]
    assert list(out["expected_answer"]) == ["v=1", "v=1"]



import sempy_labs as labs
from .base import BaseSource


class LakehouseSource(BaseSource):
    """Lakehouse implementation of BaseSource."""

    def connect(self):
        # inherits artifact_id_or_name from BaseSource
        return labs.ConnectLakehouse(
            lakehouse=self.artifact_id_or_name, workspace=self.workspace_id_or_name
        )


from uuid import UUID
import typing as t
import sempy.fabric as sf
from sempy.fabric.exceptions import FabricHTTPException
from typing import Sequence

_DEFAULT_ITEM_TYPES: tuple[str, ...] = (
    "Lakehouse",
    "Warehouse",
    "KQLDatabase",
    "SemanticModel",
)
def get_artifact_by_id_or_name(
    identifier: str | UUID,
    type_or_types: str | Sequence[str] | None,
    workspace_id: t.Optional[UUID | str] = None,
) -> t.Tuple[str, UUID]:
    """
    Retrieve the artifact name and ID by its name or ID within the specified workspace.
    Supports a single artifact type or a list of types.

    Parameters
    ----------
    identifier : str or UUID
        The name or ID of the artifact.
    type_or_types : str or list of str
        The type(s) of the artifact to search (e.g., "Lakehouse" or ["Lakehouse", "DataAgent"]).
    workspace_id : Optional[str or UUID], optional
        The workspace ID. If not provided, the current workspace ID is used.

    Returns
    -------
    tuple[str, UUID]
        A tuple containing the artifact name and its UUID.

    Raises
    ------
    ValueError
        If the workspace contains no artifacts of the supplied type(s),
        or if the specified artifact is not found.
    KeyError
        If required columns are missing in the artifact DataFrame.
    """

    # normalize parameter
    if type_or_types is None:
        types: Sequence[str] = _DEFAULT_ITEM_TYPES
    elif isinstance(type_or_types, str):
        types = [type_or_types]
    else:
        types = type_or_types

    last_error = None

    for artifact_type in types:
        try:
            df = sf.list_items(type=artifact_type, workspace=workspace_id)
            if df.empty:
                continue

            if isinstance(identifier, UUID):
                artifact_row = df[df["Id"] == str(identifier)]
                if not artifact_row.empty:
                    artifact_name = artifact_row["Display Name"].values[0]
                    return artifact_name, identifier
            else:
                artifact_row = df[df["Display Name"] == identifier]
                if not artifact_row.empty:
                    artifact_id = UUID(artifact_row["Id"].values[0])
                    return identifier, artifact_id

        except (ValueError, KeyError) as e:
            last_error = e
            continue

    if last_error is None:
        raise ValueError(
            f"Workspace contains no artifacts of the supplied type(s): {types}"
        )
    else:
        raise ValueError(
            f"Artifact '{identifier}' not found in any of the types: {types}"
        ) from last_error


def resolve_workspace_name_and_id(workspace: str | UUID | None) -> t.Tuple[str, UUID]:
    """
    Resolve the workspace name and ID based on the provided input.

    Parameters
    ----------
    workspace : str or UUID or None
        The workspace name or ID. If None, the current notebook's workspace is used.

    Returns
    -------
    tuple[str, UUID]
        A tuple containing the workspace name and its UUID.
    """
    if workspace is None:
        workspace_id = UUID(sf.get_notebook_workspace_id())
        workspace_name = sf.resolve_workspace_name(workspace_id)
    else:
        workspace_name = sf.resolve_workspace_name(workspace)
        workspace_id = UUID(sf.resolve_workspace_id(workspace))

    return workspace_name, workspace_id


def get_workspace_capacity_id(workspace_id: str | UUID) -> str:
    """
    Retrieve the capacity ID for the specified workspace.

    Parameters
    ----------
    workspace_id : str or UUID
        The ID of the workspace.

    Returns
    -------
    str
        The capacity ID associated with the workspace.

    Raises
    ------
    FabricHTTPException
        If the workspace data retrieval fails.
    ValueError
        If the capacity ID is not found in the response.
    """
    fabric_rest_client = sf.FabricRestClient()
    response = fabric_rest_client.get(f'/v1/workspaces/{workspace_id}')

    if response.status_code != 200:
        raise FabricHTTPException(
            f"Failed to retrieve workspace data: {response.status_code}, {response.text}"
        )

    content = response.json()
    capacity_id = content.get('capacityId')
    if not capacity_id:
        raise ValueError("Capacity ID not found in the response.")

    return capacity_id



import pandas as pd
from typing import Any, Dict, Optional
from azure.kusto.data import (
    ClientRequestProperties,
    KustoClient,
    KustoConnectionStringBuilder,
)
from azure.kusto.data.helpers import dataframe_from_result_table
from azure.kusto.data.exceptions import KustoServiceError, KustoClientError
from synapse.ml.fabric.token_utils import TokenUtils

from .base import BaseSource

# One TokenUtils instance per module → avoids recreating it for every token request
_token_utils = TokenUtils()

class _KustoConnection:
    """Lightweight wrapper that executes readonly KQL queries."""

    def __init__(self, cluster: str, database: str):
        kcsb = KustoConnectionStringBuilder.with_token_provider(
            cluster,
            lambda: _token_utils.get_access_token("kusto", _token_utils.get_aad_token()),
        )
        self._client = KustoClient(kcsb)
        self._database = database

    def query(
        self,
        kql: str,
        *,
        max_rows: int = 100,
        request_options: Optional[Dict[str, Any]] = None,
    ) -> pd.DataFrame:
        """Run a readonly KQL query and return a DataFrame.

        request_options – optional extra options passed to ClientRequestProperties.
        """
        props = ClientRequestProperties()
        props.set_option("query_take_max_records", max_rows)
        props.set_option("truncation_max_records", max_rows)
        props.set_option("request_readonly", True)

        # allow callers to add/override options
        if request_options:
            for k, v in request_options.items():
                props.set_option(k, v)

        try:
            response = self._client.execute(self._database, kql, properties=props)
            return dataframe_from_result_table(response.primary_results[0])
        except KustoServiceError as kse:
            raise RuntimeError(f"Kusto service error: {kse}") from kse
        except KustoClientError as kce:
            raise RuntimeError(f"Kusto client error: {kce}") from kce
        except Exception as exc:
            raise RuntimeError(f"Unexpected Kusto query failure: {exc}") from exc
        
    


    # Context-management
    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc, tb):
        # KustoClient currently has no public close(), but handle future SDKs safely.
        close_fn = getattr(self._client, "close", None)
        if callable(close_fn):
            close_fn()
        return False  


class KustoDBSource(BaseSource):
    """Kusto (Azure Data Explorer) implementation of BaseSource."""
    def __init__(self, cfg: dict[str, Any]) -> None:
        # store full dict locally
        self._cfg = cfg
        # promote the fields that BaseSource expects
        super().__init__(
            artifact_id_or_name=cfg.get("database_name") or cfg.get("display_name"),
            workspace_id_or_name=cfg.get("workspace_id"),
        )

    def connect(self):
        # Prefer explicit endpoint (full URI). Fallback to workspace GUID.
        cfg = self._cfg                       # full JSON
        cluster = cfg.get("endpoint") or cfg["workspace_id"]
        if not str(cluster).startswith(("http://", "https://")):
            cluster = f"https://{cluster}"
        return _KustoConnection(cluster=cluster, database=cfg["database_name"])


import pytest
import unittest
from unittest.mock import MagicMock
from fabric.dataagent.client._fabric_data_agent_mgmt import Datasource
from fabric.dataagent.client._fabric_data_agent_api import FabricDataAgentAPI
from fabric.dataagent.client._tagged_value import TaggedValue


class TestDatasource(unittest.TestCase):
    def setUp(self):
        self.mock_client = MagicMock(spec=FabricDataAgentAPI)
        self.datasource_id = "test_datasource_id"
        self.datasource = Datasource(self.mock_client, self.datasource_id)

    def test_init(self):
        self.assertEqual(self.datasource._client, self.mock_client)
        self.assertEqual(self.datasource._id, self.datasource_id)

    def test_repr(self):
        expected_repr = f"Datasource({self.datasource_id})"
        self.assertEqual(repr(self.datasource), expected_repr)

    def test_pretty_print(self):
        self.mock_client.get_datasource.return_value = TaggedValue(
            {
                "display_name": "test_datasource",
                "elements": [
                    {
                        "display_name": "Element1",
                        "type": "semantic_model.table",
                        "is_selected": True,
                        "children": [],
                    },
                    {
                        "display_name": "Element2",
                        "type": "semantic_model.table",
                        "is_selected": False,
                        "children": [],
                    },
                ],
            },
            "etag",
        )
        self.datasource.pretty_print(include_type=True)

    def test_get_fewshots(self):
        self.mock_client.get_datasource_fewshots.return_value = TaggedValue(
            {
                "fewShots": [
                    {
                        "id": "1",
                        "question": "Q1",
                        "query": "Query1",
                        "state": "State1",
                        "embedding": "Embedding1",
                    },
                    {
                        "id": "2",
                        "question": "Q2",
                        "query": "Query2",
                        "state": "State2",
                        "embedding": "Embedding2",
                    },
                ]
            },
            "etag",
        )
        fewshots = self.datasource.get_fewshots()
        self.assertEqual(len(fewshots), 2)

    def test_add_fewshots(self):
        self.mock_client.get_datasource_fewshots.return_value = TaggedValue(
            {"fewShots": []}, "etag"
        )
        fewshot_id = self.datasource.add_fewshots({"Q1": "Query1"})

        self.mock_client.set_datasource_fewshots.assert_called_once()

    def test_remove_fewshot(self):
        self.mock_client.get_datasource_fewshots.return_value = TaggedValue(
            {
                "fewShots": [
                    {
                        "id": "1",
                        "question": "Q1",
                        "query": "Query1",
                        "state": "State1",
                        "embedding": "Embedding1",
                    }
                ]
            },
            "etag",
        )
        self.datasource.remove_fewshot("1")
        self.mock_client.set_datasource_fewshots.assert_called_once()

    def test_successful_update_configuration(self):
        self.mock_client.get_datasource.return_value = TaggedValue(
            {"type": "lakehouse_tables"}, "etag"
        )
        self.datasource.update_configuration(
            instructions="New instructions",
            schema_mode="New schema mode",
            user_description="New description",
        )
        self.mock_client.set_datasource.assert_called_once()

    def test_failing_update_configuration(self):
        self.mock_client.get_datasource.return_value = TaggedValue(
            {"type": "semantic_model"}, "etag"
        )
        with pytest.raises(ValueError):
            self.datasource.update_configuration(
                instructions="New instructions",
                schema_mode="New schema mode",
                user_description="New description",
            )

    def test_select0(self):
        # Case 1: 0 items selected
        self.mock_client.get_datasource.return_value = TaggedValue(
            {
                "id": self.datasource_id,
                "elements": [
                    {
                        "display_name": "Element1",
                        "is_selected": False,
                        "type": "semantic_model.table",
                        "children": [],
                    }
                ],
            },
            "etag",
        )
        self.datasource.select("Element1")
        self.mock_client.set_datasource.assert_called_once()

    def test_select1(self):
        # Case 2: Path with 3 levels
        self.mock_client.get_datasource.return_value = TaggedValue(
            {
                "id": self.datasource_id,
                "elements": [
                    {
                        "display_name": "Level1",
                        "is_selected": False,
                        "children": [
                            {
                                "display_name": "Level2",
                                "is_selected": False,
                                "children": [
                                    {
                                        "display_name": "Level3",
                                        "type": "warehouse_tables.table",
                                        "is_selected": False,
                                        "children": [],
                                    }
                                ],
                            }
                        ],
                    }
                ],
            },
            "etag",
        )
        self.datasource.select("Level1", "Level2", "Level3")

        # Extract the actual call arguments
        actual_args = self.mock_client.set_datasource.call_args[0][0].value

        expected_args = {
            "id": self.datasource_id,
            "elements": [
                {
                    "display_name": "Level1",
                    "is_selected": False,
                    "children": [
                        {
                            "display_name": "Level2",
                            "is_selected": False,
                            "children": [
                                {
                                    "display_name": "Level3",
                                    "type": "warehouse_tables.table",
                                    "is_selected": True,
                                    "children": [],
                                }
                            ],
                        }
                    ],
                }
            ],
        }

        self.assertEqual(actual_args, expected_args)

    def test_unselect(self):
        self.mock_client.get_datasource.return_value = TaggedValue(
            {
                "elements": [
                    {
                        "display_name": "Element1",
                        "is_selected": True,
                        "children": [],
                        "type": "lakehouse_tables.table",
                    }
                ]
            },
            "etag",
        )
        self.datasource.unselect("Element1")
        self.mock_client.set_datasource.assert_called_once()

    def test_select_parent(self):
        # Case: Selecting a parent element should mark all child elements as selected
        self.mock_client.get_datasource.return_value = TaggedValue(
            {
                "id": self.datasource_id,
                "elements": [
                    {
                        "display_name": "dbo",
                        "is_selected": False,
                        "children": [
                            {
                                "display_name": "table1",
                                "type": "kusto.table",
                                "is_selected": False,
                                "children": [
                                    {
                                        "display_name": "col1",
                                        "is_selected": True,
                                        "children": [],
                                    },
                                    {
                                        "display_name": "col2",
                                        "is_selected": False,
                                        "children": [],
                                    },
                                    {
                                        "display_name": "col3",
                                        "is_selected": False,
                                        "children": [],
                                    },
                                ],
                            }
                        ],
                    }
                ],
            },
            "etag",
        )
        self.datasource.select("dbo", "table1")

        # Extract the actual call arguments
        actual_args = self.mock_client.set_datasource.call_args[0][0].value

        expected_args = {
            "id": self.datasource_id,
            "elements": [
                {
                    "display_name": "dbo",
                    "is_selected": False,
                    "children": [
                        {
                            "display_name": "table1",
                            "type": "kusto.table",
                            "is_selected": True,
                            "children": [
                                {
                                    "display_name": "col1",
                                    "is_selected": True,
                                    "children": [],
                                },
                                {
                                    "display_name": "col2",
                                    "is_selected": False,
                                    "children": [],
                                },
                                {
                                    "display_name": "col3",
                                    "is_selected": False,
                                    "children": [],
                                },
                            ],
                        }
                    ],
                }
            ],
        }

        self.assertEqual(actual_args, expected_args)

    def test_select_many(self):
        # Case: Selecting a parent element should mark all child elements as selected
        self.mock_client.get_datasource.return_value = TaggedValue(
            {
                "id": self.datasource_id,
                "elements": [
                    {
                        "display_name": "dbo",
                        "is_selected": False,
                        "children": [
                            {
                                "display_name": "table1",
                                "type": "kusto.table",
                                "is_selected": False,
                                "children": [
                                    {
                                        "display_name": "col1",
                                        "is_selected": False,
                                        "children": [],
                                    },
                                ],
                            },
                            {
                                "display_name": "table2",
                                "type": "kusto.table",
                                "is_selected": False,
                                "children": [
                                    {
                                        "display_name": "col2",
                                        "is_selected": False,
                                        "children": [],
                                    },
                                ],
                            },
                            {
                                "display_name": "table3",
                                "type": "kusto.table",
                                "is_selected": False,
                                "children": [
                                    {
                                        "display_name": "col3",
                                        "is_selected": False,
                                        "children": [],
                                    },
                                ],
                            }
                        ],
                    }
                ],
            },
            "etag",
        )
        self.datasource.select(["dbo", "table1"], ["dbo", "table3"])

        # Extract the actual call arguments
        actual_args = self.mock_client.set_datasource.call_args[0][0].value

        expected_args = {
            "id": self.datasource_id,
            "elements": [
                {
                    "display_name": "dbo",
                    "is_selected": False,
                    "children": [
                        {
                            "display_name": "table1",
                            "type": "kusto.table",
                            "is_selected": True,
                            "children": [
                                {
                                    "display_name": "col1",
                                    "is_selected": False,
                                    "children": [],
                                },
                            ],
                        },
                        {
                            "display_name": "table2",
                            "type": "kusto.table",
                            "is_selected": False,
                            "children": [
                                {
                                    "display_name": "col2",
                                    "is_selected": False,
                                    "children": [],
                                },
                            ],
                        },
                        {
                            "display_name": "table3",
                            "type": "kusto.table",
                            "is_selected": True,
                            "children": [
                                {
                                    "display_name": "col3",
                                    "is_selected": False,
                                    "children": [],
                                },
                            ],
                        }
                    ],
                }
            ],
        }

        self.assertEqual(actual_args, expected_args)

    def test_select_column(self):
        # Case 1: 0 items selected
        self.mock_client.get_datasource.return_value = TaggedValue(
            {
                "id": self.datasource_id,
                "elements": [
                    {
                        "display_name": "Element1",
                        "is_selected": False,
                        "type": "semantic_model.column",
                        "children": [],
                    }
                ],
            },
            "etag",
        )

        with pytest.raises(ValueError):
            self.datasource.select("Element1")

    def test_successful_update_description(self):
        self.mock_client.get_datasource.return_value = TaggedValue(
            {
                "type": "lakehouse_tables",
                "elements": [
                    {
                        "display_name": "dbo",
                        "type": "lakehouse_tables.schema",
                        "children": [
                            {
                                "display_name": "table1",
                                "type": "lakehouse_tables.table",
                                "description": None,
                                "children": [
                                    {
                                        "display_name": "col1",
                                        "type": "lakehouse_tables.column",
                                        "description": None,
                                        "children": [],
                                    },
                                ],
                            }
                        ],
                    }
                ],
            },
            "etag",
        )

        self.datasource.update_description(
            {
                ("dbo", "table1"): "table1 description",
                ("dbo", "table1", "col1"): "col1 description",
            }
        )

        # Check that set_datasource was called and descriptions were updated
        args = self.mock_client.set_datasource.call_args[0][0].value
        table = args["elements"][0]["children"][0]
        col = table["children"][0]
        self.assertEqual(table["description"], "table1 description")
        self.assertEqual(col["description"], "col1 description")

    def test_update_description_unsupported_datasource_type(self):
        self.mock_client.get_datasource.return_value = TaggedValue(
            {
                "type": "semantic_model",
                "elements": [],
            },
            "etag",
        )
        with pytest.raises(ValueError):
            self.datasource.update_description({("dbo", "table1"): "desc"})

    def test_update_description_path_not_found(self):
        self.mock_client.get_datasource.return_value = TaggedValue(
            {
                "type": "lakehouse_tables",
                "elements": [
                    {
                        "display_name": "dbo",
                        "type": "lakehouse_tables.schema",
                        "children": [
                            {
                                "display_name": "table1",
                                "type": "lakehouse_tables.table",
                                "description": None,
                                "children": [],
                            }
                        ],
                    }
                ],
            },
            "etag",
        )
        with pytest.raises(ValueError):
            self.datasource.update_description({("dbo", "table2"): "desc"})

    def test_update_description_unsupported_datasource(self):
        self.mock_client.get_datasource.return_value = TaggedValue(
            {
                "type": "kusto",
                "elements": [],
            },
            "etag",
        )
        with pytest.raises(ValueError):
            self.datasource.update_description({("dbo", "table1"): "desc"})

    def test_update_description_wrong_type(self):
        self.mock_client.get_datasource.return_value = TaggedValue(
            {
                "type": "lakehouse_tables",
                "elements": [
                    {
                        "display_name": "dbo",
                        "type": "lakehouse_tables.schema",
                        "children": [
                            {
                                "display_name": "file1",
                                "type": "lakehouse_files.directory",
                                "description": None,
                                "children": [],
                            }
                        ],
                    }
                ],
            },
            "etag",
        )
        with pytest.raises(ValueError):
            self.datasource.update_description({("dbo", "file1"): "desc"})



import sempy_labs as labs
from .base import BaseSource


class WarehouseSource(BaseSource):
    """Data-warehouse implementation of BaseSource."""

    def connect(self):
        return labs.ConnectWarehouse(
            warehouse=self.artifact_id_or_name, workspace=self.workspace_id_or_name
        )



import uuid
from typing import Iterator, Optional
from uuid import UUID

import sempy.fabric as fabric
from sempy.fabric.exceptions import FabricHTTPException
import synapse.ml.fabric.service_discovery as service_discovery
from ._tagged_value import TaggedValue
from ._util import (
    get_artifact_by_id_or_name,
    get_workspace_capacity_id,
    resolve_workspace_name_and_id,
)


class FabricDataAgentAPI:
    """
    Client for interacting with the Fabric DataAgent REST API.

    Attributes
    ----------
    _data_agent_base_url : str
        The base URL for DataAgent management endpoints.
    _client : FabricRestClient
        The REST client used for API communication.
    _workspace_id : str
        The ID of the workspace associated with the DataAgent.
    _workspace_name : str
        The name of the workspace associated with the DataAgent.
    data_agent_name : str
        The name of the DataAgent.
    data_agent_id : UUID
        The unique identifier of the DataAgent.
    capacity_id : str
        The capacity ID associated with the workspace.
    workload_resource_moniker : str
        A unique identifier for the workload resource.
    scenario : str
        The scenario type, default is "data_agent".
    stage : str
        The stage type, default is "sandbox".
    """

    def __init__(
        self, data_agent: str | UUID, workspace: Optional[str | UUID] = None
    ) -> None:
        """
        Initialize the client for interacting with the Fabric REST API.

        Parameters
        ----------
        data_agent : str or UUID
            The name or ID of the Data Agent to be used.
        workspace : Optional[str or UUID], optional
            The workspace name or ID. If not provided, it will be fetched automatically.
        """
        self._client = fabric.FabricRestClient()

        self.workspace_name, self.workspace_id = resolve_workspace_name_and_id(
            workspace
        )

        self.scenario = "data_agent"
        self.stage = "sandbox"
        self.workload_resource_moniker = str(uuid.uuid4())  # Generate a new GUID

        # resolve the Data Agent artifact ID and capacity ID
        self.data_agent_name, self.data_agent_id = get_artifact_by_id_or_name(
            data_agent, ["DataAgent", "AISkill"], self.workspace_id
        )

        self.capacity_id = get_workspace_capacity_id(self.workspace_id)

        # Get the fabric environment configuration
        self._host = service_discovery.get_fabric_env_config().fabric_env_config.wl_host
        self._data_agent_base_url = f"{self._host}/webapi/capacities/{self.capacity_id}/workloads/ML/AISkill/Automatic/workspaces/{self.workspace_id}/dataagents/{self.data_agent_id}"

    def get_configuration(self) -> TaggedValue:
        """
        Retrieve the configuration from the Data Agent.

        Returns
        -------
        TaggedValue
            The configuration of the Data Agent.
        """

        return TaggedValue.from_response(
            self._client.get(f"{self._data_agent_base_url}/management/configuration")
        )

    def set_configuration(self, value: TaggedValue) -> TaggedValue:
        """
        Set the configuration for the Data Agent.

        Parameters
        ----------
        value : TaggedValue
            The configuration to set.

        Returns
        -------
        TaggedValue
            The updated configuration of the Data Agent with the new ETag.
        """

        # copy
        body = dict(value.value)

        # we can't update dataSources directly
        if "dataSources" in body:
            del body["dataSources"]

        response = self._client.patch(
            f"{self._data_agent_base_url}/management/configuration",
            json=body,
            headers={
                "x-ms-ai-aiskill-stage": self.stage,
                "If-Match": value.etag,
            },
        )

        if response.status_code != 200:
            raise FabricHTTPException(response)

        return TaggedValue(value.value, response.headers.get("ETag", ""))

    def get_datasources(self) -> Iterator[TaggedValue]:
        """
        Retrieve the list of datasources from the Data Agent.

        Returns
        -------
        Iterator[TaggedValue]
            The list of datasources.
        """

        config = self.get_configuration()

        for ds in config.value.get("dataSources", []):
            yield self.get_datasource(ds["id"])

    def get_datasource(self, datasource_id: str) -> TaggedValue:
        """
        Retrieve a specific datasource by its ID.

        Parameters
        ----------
        datasource_id : str
            The ID of the datasource.

        Returns
        -------
        TaggedValue
            The datasource.
        """
        return TaggedValue.from_response(
            self._client.get(
                f"{self._data_agent_base_url}/management/datasources/{datasource_id}"
            )
        )

    def remove_datasource(self, datasource_id: str, configuration_etag: str) -> None:
        """
        Remove a datasource from the Data Agent.

        Parameters
        ----------
        datasource_id : str
            The ID of the datasource to remove.
        configuration_etag : str
            The ETag of the current configuration.

        Raises
        ------
        FabricHTTPException
            If the response status code is not 200.
        """
        response = self._client.delete(
            f"{self._data_agent_base_url}/management/datasources/{datasource_id}",
            headers={
                "x-ms-ai-aiskill-stage": self.stage,
                "If-Match": configuration_etag,
            },
        )

        if response.status_code != 200:
            raise FabricHTTPException(response)

    def set_datasource(self, datasource: TaggedValue) -> TaggedValue:
        """
        Set the datasource for the Data Agent.

        Parameters
        ----------
        datasource : TaggedValue
            The datasource to set.

        Returns
        -------
        TaggedValue
            The updated datasource of the Data Agent with the new ETag.
        """

        response = self._client.put(
            f"{self._data_agent_base_url}/management/datasources/{datasource.value['id']}",
            json=datasource.value,
            headers={
                "x-ms-ai-aiskill-stage": self.stage,
                "If-Match": datasource.etag,
            },
        )

        if response.status_code != 200:
            raise FabricHTTPException(response)

        return TaggedValue(datasource.value, response.headers.get("ETag", ""))

    def get_publishing_info(self) -> TaggedValue:
        """
        Retrieve the publishing information.

        Returns
        -------
        TaggedValue
            The publishing information.
        """

        return TaggedValue.from_response(
            self._client.get(f"{self._data_agent_base_url}/management/publishing")
        )

    def put_publish_info(self, value: TaggedValue) -> TaggedValue:
        """
        Update the publishing information with the provided description.

        Parameters
        ----------
        value : TaggedValue
            The new publishing information.

        Returns
        -------
        TaggedValue
            The updated publishing information with the new ETag.
        """

        return TaggedValue.from_response(
            self._client.put(
                f"{self._data_agent_base_url}/management/publishing",
                json=value.value,
                headers={
                    "Content-Type": "application/json",
                    "If-Match": value.etag,
                },
            )
        )

    def add_datasource(self, datasource_body: dict) -> dict:
        """
        Add a new datasource to the system.

        Parameters
        ----------
        datasource_body : dict
            A dictionary containing the details of the datasource to be added.

        Returns
        -------
        dict
            A dictionary containing the response content of the added datasource.

        Raises
        ------
        FabricHTTPException
            If the response status code is not 200.
        """

        # Common header for all the datasources
        headers = {
            "Content-Type": "application/json",
            "x-ms-workload-resource-moniker": f"{self.workload_resource_moniker}",
            "x-ms-ai-assistant-scenario": f"{self.scenario}",
            "x-ms-ai-aiskill-stage": f"{self.stage}",
        }
        # Make the POST request
        add_datasource_response = self._client.post(
            f"{self._data_agent_base_url}/management/datasources",
            json=datasource_body,
            headers=headers,
        )
        # Raise FabricHTTPException if the status code is not 200
        if add_datasource_response.status_code != 200:
            raise FabricHTTPException(add_datasource_response)

        return add_datasource_response.json()

    def publish(self, configuration: TaggedValue) -> None:
        """
        Publish the Data Agent configuration.

        Parameters
        ----------
        configuration : TaggedValue
            The configuration to publish.

        Raises
        ------
        FabricHTTPException
            If the response status code is not 200.
        """
        response = self._client.put(
            f"{self._data_agent_base_url}/management/deploy",
            headers={
                "Content-Type": "application/json",
                "If-Match": configuration.etag,
            },
        )

        if response.status_code != 200:
            raise FabricHTTPException(response)

    def get_schema(
        self,
        datasource_workspace_id: str,
        datasource_id: str,
        type: str,
        response_source: str = "live",
    ) -> dict:
        """
        Retrieve the schema for a specified datasource.

        Parameters
        ----------
        datasource_workspace_id : UUID
            The workspace ID of the datasource.
        datasource_id : str
            The unique identifier of the datasource.
        type : str
            The type of the datasource (e.g., "LakehouseTables", "Kusto").
        response_source : str, optional
            The source of the response, default is "live".

        Returns
        -------
        dict
            The JSON response containing the schema of the datasource.

        Raises
        ------
        FabricHTTPException
            If the response status code is not 200.
        """

        # Construct the URL
        get_schema_url = f"{self._host}/webapi/capacities/{self.capacity_id}/workloads/ML/AISkill/Automatic/v1/workspaces/{datasource_workspace_id}/artifacts/{datasource_id}/schema?responseSource={response_source}&dataSourceType={type}"

        # Construct the headers
        headers = {
            "Content-Type": "application/json",
            "x-ms-upstream-artifact-id": str(self.data_agent_id),
            "x-ms-workload-resource-moniker": self.workload_resource_moniker,
        }
        # Make the GET request
        response = self._client.get(get_schema_url, headers=headers)
        # Raise FabricHTTPException if the status code is not 200
        if response.status_code != 200:
            raise FabricHTTPException(response)
        return response.json()

    def get_datasource_fewshots(self, datasource_id: str):
        return TaggedValue.from_response(
            self._client.get(
                f"{self._data_agent_base_url}/management/datasources/{datasource_id}/fewshots"
            )
        )

    def set_datasource_fewshots(self, datasource_id: str, fewshots: TaggedValue):
        response = self._client.post(
            f"{self._data_agent_base_url}/management/datasources/{datasource_id}/fewshots",
            json=fewshots.value,
            headers={"If-Match": fewshots.etag},
        )

        if response.status_code != 200:
            raise FabricHTTPException(response)



import unittest
from unittest.mock import patch, MagicMock
from fabric.dataagent.client._fabric_openai import FabricOpenAI


class TestFabricOpenAI(unittest.TestCase):

    @patch('fabric.dataagent.client._fabric_openai.get_fabric_env_config')
    @patch('fabric.dataagent.client._fabric_openai.get_artifact_by_id_or_name')
    @patch('fabric.dataagent.client._fabric_openai.get_workspace_capacity_id')
    @patch('fabric.dataagent.client._fabric_openai.resolve_workspace_name_and_id')
    def setUp(
        self,
        mock_resolve_workspace,
        mock_get_capacity,
        mock_get_artifact,
        mock_get_config,
    ):

        self.mock_resolve_workspace = mock_resolve_workspace
        self.mock_get_capacity = mock_get_capacity
        self.mock_get_artifact = mock_get_artifact
        self.mock_get_config = mock_get_config

        self.artifact_name = "artifact_name"
        self.artifact_id = "artifact_id"
        self.workspace_name = "workspace_name"
        self.workspace_id = "workspace_id"
        self.capacity_id = "capacity_id"
        self.wl_host = "https://fabric_host"

        self.mock_resolve_workspace.return_value = (
            self.workspace_name,
            self.workspace_id,
        )
        self.mock_get_capacity.return_value = self.capacity_id
        self.mock_get_artifact.return_value = (self.artifact_name, self.artifact_id)
        self.mock_get_config.return_value = MagicMock(
            fabric_env_config=MagicMock(wl_host=self.wl_host)
        )

        self.fabric_openai = FabricOpenAI(artifact_name=self.artifact_name)

    def test_initialization(self):
        # Assert the instance variables
        self.assertEqual(self.fabric_openai.workspace_name, self.workspace_name)
        self.assertEqual(self.fabric_openai.workspace_id, self.workspace_id)
        self.assertEqual(self.fabric_openai.capacity_id, self.capacity_id)
        self.assertEqual(self.fabric_openai.artifact_name, self.artifact_name)
        self.assertEqual(self.fabric_openai.artifact_id, self.artifact_id)

        # Assert the default values
        self.assertEqual(self.fabric_openai.api_version, "2024-05-01-preview")
        self.assertEqual(self.fabric_openai.assistant_scenario, "aiskill")
        self.assertEqual(self.fabric_openai.ai_skill_stage, "sandbox")
        self.assertTrue(self.fabric_openai._moniker_id)

        # Assert the base_url
        # TODO: Change once base_url is modified:
        expected_base_url = f"{self.wl_host}/webapi/capacities/{self.capacity_id}/workloads/ML/AISkill/Automatic/v1/workspaces/{self.workspace_id}/artifacts/{self.artifact_id}/aiassistant/openai/"
        self.assertEqual(self.fabric_openai.base_url, expected_base_url)

    @patch('fabric.dataagent.client._fabric_openai.SynapseTokenProvider')
    def test_prepare_options(self, mock_token_provider):

        mock_token_provider.return_value = MagicMock(return_value="mock_token")
        options = MagicMock(headers={"Custom-Header": "value"})

        prepared_options = self.fabric_openai._prepare_options(options)

        # Assert the headers
        self.assertEqual(prepared_options.headers["Authorization"], "Bearer mock_token")
        self.assertEqual(prepared_options.headers["Accept"], "application/json")
        self.assertIn("ActivityId", prepared_options.headers)
        self.assertEqual(
            prepared_options.headers["x-ms-workload-resource-moniker"],
            self.fabric_openai._moniker_id,
        )
        self.assertEqual(
            prepared_options.headers["x-ms-ai-assistant-scenario"], "aiskill"
        )
        self.assertEqual(prepared_options.headers["x-ms-ai-aiskill-stage"], "sandbox")

    def test_make_status_error(self):

        response = MagicMock(
            status_code=500,
            headers={
                "x-ms-root-activity-id": "raid_value",
                "ActivityId": "correlation_id_value",
            },
        )

        error = self.fabric_openai._make_status_error(
            "Error message", body={}, response=response
        )

        # Assert the error message
        self.assertIn("RAID: raid_value", str(error))
        self.assertIn("Correlation ID: correlation_id_value", str(error))

    @patch('sempy.fabric.FabricRestClient')
    def test_get_or_create_thread(self, MockFabricRestClient):
        self.mock_client = MockFabricRestClient.return_value
        self.mock_client.get.return_value = MagicMock(status_code=200)
        self.fabric_openai.beta.threads.retrieve = MagicMock(return_value="123")

        self.fabric_openai.get_or_create_thread("new_thread")
        self.mock_client.get.assert_called_once()

    @patch('sempy.fabric.FabricRestClient')
    def test_delete_thread(self, MockFabricRestClient):
        self.mock_client = MockFabricRestClient.return_value
        self.mock_client.delete.return_value = MagicMock(status_code=200)
        self.fabric_openai.delete_thread("new_thread")
        self.mock_client.delete.assert_called_once()



from fabric.dataagent.evaluation.evaluator_api import evaluate_data_agent, get_evaluation_details, get_evaluation_summary, get_evaluation_summary_per_question

__all__ = [
    "evaluate_data_agent",
    "get_evaluation_details",
    "get_evaluation_summary",
    "get_evaluation_summary_per_question",
]



import sempy.fabric as fabric
from typing import Optional
from uuid import UUID
from sempy.fabric.exceptions import FabricHTTPException

from ._fabric_data_agent_mgmt import FabricDataAgentManagement

ARTIFACT_TYPE = 'dataagent'

def create_data_agent(
    data_agent_name: str, workspace_id: Optional[UUID | str] = None
) -> FabricDataAgentManagement:
    """
    Create a new Data Agent and return an instance of FabricDataAgentAPI.

    Parameters
    ----------
    data_agent_name : str
        The name of the Data Agent to be created.
    workspace_id : Optional[UUID or str], optional
        The workspace ID. If not provided, it will be fetched automatically.

    Returns
    -------
    FabricDataAgentAPI
        An instance of FabricDataAgentAPI initialized with the created Data Agent.

    Raises
    ------
    FabricHTTPException
        If the response status code is not 200.
    """
    if not workspace_id:
        workspace_id = fabric.get_notebook_workspace_id()

    create_artifact_url = f"v1/workspaces/{workspace_id}/{ARTIFACT_TYPE}s"
    # Construct the body
    create_artifact_body = {
        "artifactType": "LLMPlugin",
        "displayName": data_agent_name,
    }

    fabric_client = fabric.FabricRestClient()
    response = fabric_client.post(create_artifact_url, json=create_artifact_body)

    if response.status_code not in [200, 201, 202]:
        raise FabricHTTPException(response)

    return FabricDataAgentManagement(data_agent_name, workspace_id)


def delete_data_agent(data_agent_name_or_id: str,
                      workspace_id: Optional[UUID | str] = None
                      ) -> None:
    """
    Delete a Data Agent.

    Parameters
    ----------
    data_agent_name_or_id : str
        The name or ID of the Data Agent to delete.

    Raises
    ------
    FabricHTTPException
        If the response status code is not 200.
    """
    if not workspace_id:
        workspace_id = fabric.get_notebook_workspace_id()
    if isinstance(data_agent_name_or_id, UUID):
        data_agent_id = str(data_agent_name_or_id)
    else:
        df = fabric.list_items(type="DataAgent", workspace=workspace_id)
        match = df[
            (df["Display Name"] == data_agent_name_or_id)
            | (df["Id"] == data_agent_name_or_id)
        ]

        if match.empty:
            df_ai_skills = fabric.list_items(type="AISkill")
            match = df_ai_skills[
                (df_ai_skills["Display Name"] == data_agent_name_or_id)
                | (df_ai_skills["Id"] == data_agent_name_or_id)
            ]

        if match.empty:
            raise ValueError(f"Data Agent '{data_agent_name_or_id}' not found.")

        data_agent_id = match["Id"].values[0]

    workspace_id = fabric.get_notebook_workspace_id()
    artifact_url = f"v1/workspaces/{workspace_id}/{ARTIFACT_TYPE}s/{data_agent_id}"

    fabric_client = fabric.FabricRestClient()
    response = fabric_client.delete(artifact_url)

    if response.status_code != 200:
        raise FabricHTTPException(response)



from .lakehouse import LakehouseSource
from .warehouse import WarehouseSource
from .kustodb import KustoDBSource 
from .base import BaseSource

_TYPE_MAP = {
    "lakehouse_tables": LakehouseSource,
    "data_warehouse": WarehouseSource,  # TODO: DAX support
    "kusto": KustoDBSource,
}


def make_source(cfg: dict) -> BaseSource:
    """
    Build a concrete Source object from a Fabric datasource configuration.
    `cfg` must have at least keys: 'type', 'display_name'.
    """
    if "type" not in cfg or "display_name" not in cfg:
        raise ValueError("cfg must contain 'type' and 'display_name'")

    try:
        cls = _TYPE_MAP[cfg["type"]]
    except KeyError as exc:
        raise ValueError(f"Unsupported datasource type '{cfg['type']}'") from exc

    # Kusto needs the full JSON (contains "endpoint" and "database_name")
    if cfg["type"] == "kusto":
        return cls(cfg)                               # pass everything

    # lakehouse / warehouse keep the lean constructor for now
    return cls(
        artifact_id_or_name = cfg.get("id") or cfg["display_name"],
        workspace_id_or_name=cfg.get("workspace_id") or cfg.get("workspace_name"),
    )


from uuid import UUID
from typing import Union

class BaseSource:
    """
    Common interface for all Fabric datasources.
    Sub-classes must implement `connect()` and return a context manager
    exposing a `.query(sql_or_kql)` method that yields a Pandas DataFrame.
    """

    def __init__(
        self,
        artifact_id_or_name: Union[str, UUID],
        workspace_id_or_name: str | None = None,
    ) -> None:

        self.artifact_id_or_name = artifact_id_or_name
        self.workspace_id_or_name = workspace_id_or_name

    def connect(self):
        """Return a context-manager (e.g. `labs.ConnectLakehouse`)."""
        raise NotImplementedError("Sub-classes must override connect()")



import requests
from sempy.fabric.exceptions import FabricHTTPException
import typing as t


class TaggedValue:
    """
    A value along with its ETag, used to support optimistic concurrency control.

    Attributes
    ----------
    value : dict
        The data associated with this TaggedValue.
    etag : str
        The ETag associated with this value.
    """

    value: t.Dict[str, t.Any]
    etag: str

    def __init__(self, value: t.Dict[str, t.Any], etag: str):
        """
        Initialize a TaggedValue instance.

        Parameters
        ----------
        value : dict
            The data to be stored.
        etag : str
            The ETag associated with the value.
        """
        self.value = value
        self.etag = etag

    def __str__(self):
        """
        Return a string representation of the TaggedValue.

        Returns
        -------
        str
            The string representation including the value and ETag.
        """
        return f"{self.value} (ETag: {self.etag})"

    def __repr__(self):
        """
        Return the official string representation of the TaggedValue.

        Returns
        -------
        str
            The string representation suitable for debugging.
        """
        return f"TaggedValue(value={self.value}, etag={self.etag})"

    @staticmethod
    def from_response(response: requests.Response) -> "TaggedValue":
        """
        Create a TaggedValue instance from an HTTP response.

        Parameters
        ----------
        response : requests.Response
            The HTTP response containing the data and ETag.

        Returns
        -------
        TaggedValue
            An instance of TaggedValue with data and ETag extracted from the response.

        Raises
        ------
        FabricHTTPException
            If the response status code is not 200.
        """
        if response.status_code != 200:
            raise FabricHTTPException(response)

        return TaggedValue(response.json(), response.headers.get("ETag", ""))



import pytest


@pytest.mark.skip(reason="Fabric is not available")
def test_integration_fabric_data_agent_client(fabric_info):
    from fabric.dataagent.client import (
        FabricOpenAI,
        Fabricdata_agentAPI,
        Fabricdata_agentManagement,
    )
    from sempy.fabric import list_items

    # oai = FabricOpenAI("20250113-dax-1")

    # t = oai.beta.threads.create()
    # print(f"thread id: {t.id}")

    # client_api = Fabricdata_agentAPI(fabric_info["workspace_id"])
    # client = Fabricdata_agentAPI("20250113-dax-1")

    mgmt = Fabricdata_agentManagement("20250113-dax-1")

    print(mgmt.get_configuration())

    datasources = mgmt.get_datasources()

    # datasources[1].pretty_print()
    datasources[0].pretty_print()

    datasources[0].unselect("Platforms")

    datasources[0].pretty_print()
    # print(datasources[0].get_fewshots())

    # datasources[1].remove_fewshot("736359b3-3ebd-4d42-9e06-30e76ded3569")
    # print(datasources[1].get_fewshots())

    # datasources[1].add_fewshot("test5", "SELECT * from dbo.factfinance")

    # print(datasources[1].get_fewshots())

    # ds = mgmt.add_datasource("AdventureWorksLH", "ML-SHARED-data_agentV2-DATASOURCES", type="lakehouse")

    # fw = ds.get_fewshots()

    # ds.unselect()
    # ds.select("dbo", "factfinance")

    # ds.pretty_print()

    # mgmt.remove_datasource("AdventureWorksLH")

    # config = client.get_configuration()

    # print(config)
    # # c2 = client.set_configuration(config)
    # # c3 = client.set_configuration(c2)
    # # print(c3)

    # datasources = client.get_datasources()

    # print(datasources)

    # datasources[0].value["additional_instructions"] = "Make sure each column is prefixed with YY_"

    # # datasources[0] = client.set_datasource(datasources[0])

    # print(client.get_schema("9c5aa981-58eb-4844-954b-8546e137dfa3", "eee0500b-0616-4f25-8d80-faf0e2813658", "Kusto"))

    # # client2 = FabricA()

    # # client2.update_configuration(instructions="", userDescription="")

    # # client2.deploy() # deploy the AI skill (sandbox -> production)

    # # datasources = client2.get_datasources()

    # # calls are effective immediately

    # # datasources[0].pretty_print() ## tree view in notebook
    # # datasources[0].update_configuration(instructions="Make sure each column is prefixed with YY_", schema_mode="schema", user_description="7th Generation Video Games")
    # # datasources[0].select(["Games", "Id"])
    # # datasources[0].select(["dbo", "Games", "Name"])
    # # datasources[0].unselect("Games")

    # # automatically add to AI skill, unselect all tables
    # # new_datasource = client2.add_datasource(artifact_name_or_id="", workspace_id_or_name="", type="Kusto"|None)



from fabric.dataagent.client._fabric_data_agent_api import (
    FabricDataAgentAPI,
    get_artifact_by_id_or_name,
)
from fabric.dataagent.client._datasource import Datasource
import typing as t
from uuid import UUID
from sempy.fabric import list_items, get_notebook_workspace_id, resolve_workspace_id
from dataclasses import dataclass


@dataclass
class DataAgentConfiguration:
    """
    Configuration settings for an DataAgent.

    Attributes
    ----------
    instructions : str or None
        Additional instructions for the DataAgent.
    user_description : str or None
        Description provided by the user for the DataAgent.
    """

    instructions: str | None
    user_description: str | None


class FabricDataAgentManagement:
    """
    Management class for DataAgent operations.

    Attributes
    ----------
    _client : FabricDataAgentAPI
        The FabricDataAgentAPI client instance.
    """

    _client: FabricDataAgentAPI

    def __init__(
        self, data_agent: str | UUID, workspace: t.Optional[str | UUID] = None
    ) -> None:
        """
        Initialize a FabricDataAgentManagement instance.

        Parameters
        ----------
        data_agent : str or UUID
            The name or ID of the DataAgent.
        workspace : str or UUID, optional
            The name or ID of the workspace. If not provided, it defaults to the current workspace.
        """
        self._client = FabricDataAgentAPI(data_agent=data_agent, workspace=workspace)

    def update_configuration(
        self, instructions: str | None = None, user_description: str | None = None
    ) -> None:
        """
        Update the configuration of the Data Agent.

        Parameters
        ----------
        instructions : str, optional
            Additional instructions for the data agent.
        user_description : str, optional
            The user description for the data agent.
        """
        config = self._client.get_configuration()

        if instructions:
            config.value["additionalInstructions"] = instructions

        if user_description:
            config.value["userDescription"] = user_description

        self._client.set_configuration(config)

    def get_configuration(self) -> DataAgentConfiguration:
        """
        Retrieve the configuration of the Data Agent.

        Returns
        -------
        DataAgentConfiguration
            The configuration of the Data Agent.
        """
        config = self._client.get_configuration()

        return DataAgentConfiguration(
            instructions=config.value.get("additionalInstructions"),
            user_description=config.value.get("userDescription"),
        )

    def publish(self) -> None:
        """
        Publish the Data Agent configuration.
        """
        self._client.publish(self._client.get_configuration())

    def get_datasources(self) -> t.List[Datasource]:
        """
        Retrieve the list of datasources for the Data Agent.

        Returns
        -------
        t.List[Datasource]
            The list of datasources.
        """
        return [
            Datasource(self._client, datasource.value['id'])
            for datasource in self._client.get_datasources()
        ]

    def add_datasource(
        self,
        artifact_name_or_id: str,
        workspace_id_or_name: str | None = None,
        type: str | None = None,
    ) -> Datasource:
        """
        Add a new datasource to the Data Agent.

        Parameters
        ----------
        artifact_name_or_id : str
            The name or ID of the artifact.
        workspace_id_or_name : str, optional
            The workspace ID or name. Defaults to None.
        type : str, optional
            The type of the artifact. Defaults to None.

        Returns
        -------
        Datasource
            The added datasource.

        Raises
        ------
        ValueError
            If the datasource artifact type is unsupported or if the artifact cannot be found.
        """
        if workspace_id_or_name is None:
            workspace_id_or_name = get_notebook_workspace_id()

        workspace_id = resolve_workspace_id(workspace_id_or_name)

        if type is None:
            # auto-resolve type
            df_items = list_items(workspace=workspace_id_or_name)

            if isinstance(artifact_name_or_id, UUID):
                df_item = df_items[df_items['Id'] == str(artifact_name_or_id)]
            else:
                df_item = df_items[df_items['Display Name'] == artifact_name_or_id]

            if df_item.empty:
                raise ValueError(
                    f"Could not find datasource artifact '{artifact_name_or_id}' in workspace '{workspace_id_or_name}'"
                )

            if df_item.shape[0] > 1:
                raise ValueError(
                    f"Found multiple artifacts with name '{artifact_name_or_id}' in workspace '{workspace_id_or_name}'"
                )

            type = df_item['Type'].iloc[0]

        (artifact_name, artifact_id) = get_artifact_by_id_or_name(
            artifact_name_or_id, type, workspace_id_or_name
        )

        schema_types = {
            "lakehouse": "lakehouse_tables",
            "kqldatabase": "kusto",
            "warehouse": "data_warehouse",
            "semanticmodel": "semantic_model",
        }

        if type not in schema_types:
            raise ValueError(f"Unsupported artifact type '{type}'")

        # also contains "cacheLastUpdatedTime" value
        schema = self._client.get_schema(
            workspace_id, artifact_id, schema_types[type.lower()]
        )

        ds = self._client.add_datasource(schema["schema"])

        return Datasource(self._client, ds["id"])

    def remove_datasource(self, datasource_name_or_id: str) -> None:
        """
        Remove a datasource from the Data Agent.

        Parameters
        ----------
        datasource_name_or_id : str
            The name or ID of the datasource to remove.

        Raises
        ------
        ValueError
            If the datasource cannot be found.
        """
        # to remove a datasource, we need to know the etag of the configuration
        config = self._client.get_configuration()

        datasources = self._client.get_datasources()

        for ds in datasources:
            if (
                ds.value['id'] == datasource_name_or_id
                or ds.value['display_name'] == datasource_name_or_id
            ):
                self._client.remove_datasource(ds.value['id'], config.etag)
                return

        raise ValueError(f"Could not find datasource '{datasource_name_or_id}'")



import os
import jwt
import hashlib
import json
from urllib.parse import quote

import requests

env_cred = {
    "prod": "AdminUser@SynapseMLFabricIntegration.onmicrosoft.com",
}

env_workload_host = {
    "daily": "pbipdailyeus2euap-daily.pbidedicated.windows.net",
    "dxt": "pbipdxtcuseuap1-dxt.pbidedicated.windows.net",
    "msit": "pbipmsitwcus23-msit.pbidedicated.windows.net",
}


env_onelake_new = {
    "edog": "onelake-int-edog.dfs.pbidedicated.windows-int.net",
    "daily": "daily-onelake.dfs.fabric.microsoft.com",
    "dxt": "dxt-onelake.dfs.fabric.microsoft.com",
    "msit": "msit-onelake.dfs.fabric.microsoft.com",
    "prod": "onelake.dfs.fabric.microsoft.com",
}


def get_bearer_token(env, username, password, audience="pbi"):
    if audience != "pbi" and audience != "storage":
        raise ValueError("Only pbi and storage audiences are supported.")

    # XMLA client id
    client_id = "cf710c6e-dfcc-4fa8-a093-d47294e44c66"
    user, tenant = username.split("@")

    if env == "edog":
        resource = "https://analysis.windows-int.net/powerbi/api"
        authority = "login.windows-ppe.net"
    else:
        resource = (
            "https://analysis.windows.net/powerbi/api"
            if audience == "pbi"
            else "https://storage.azure.com"
        )
        authority = "login.windows.net"

    login_url = f"https://{authority}/{tenant}/oauth2/token"
    payload = f'resource={quote(resource, safe="")}&client_id={client_id}&grant_type=password&username={username}&password={quote(password, safe="")}&scope=openid'
    headers = {
        "Content-Type": "application/x-www-form-urlencoded",
        "Cookie": "fpc=AlqdPsBZ3IhEkuEX2q3BHxjyrCATAQAAAAFta9sOAAAALbyONgEAAAAZbWvbDgAAAA; stsservicecookie=estsppe",
    }
    response = requests.request("GET", login_url, headers=headers, data=payload).json()
    try:
        token_type = response["token_type"]
    except Exception as e:
        print(f"Bad response: {response}")
        raise e

    if token_type != "Bearer":
        raise ValueError("The token received is not a bearer.")

    return response["access_token"]


def get_username_from_bearer_token(bearer_token: str):
    payload = jwt.decode(bearer_token, options={"verify_signature": False})
    return payload.get("upn")


class FabricAuthUtil:
    """
    # check if .fabric-token exists
    #    parse and done :)
    #
    # 1. get corpnet user token via "az login --use-device-code" to access keyvault
    # 2. get environment (edog/dxt/msit) username/password from keyvault
    # 3. get bearer token for environment username/password
    # 4. create workspace (re-use if exists)
    # 5. store the token in .fabric-token
    """

    env_bearer_token: dict[str, str] = dict()

    user_keyvaults = {
        "prod": {
            "vault": "SynapseMLFabricTestCreds",
            "secret-name": "AdminUser-SynapseMLFabricIntegration",
        }
    }

    def __init__(self, env: str):
        self.env = env

    def get_dev_user(self):
        """Returns the developer's username."""

        return get_username_from_bearer_token(self.get_env_bearer_token())

    def get_env_username(self) -> str:
        """Returns the username for the environment (edog/dxt/msit)."""
        return env_cred[self.env]

    def get_env_password(self):
        return os.environ.get("ut_password")

    def get_env_bearer_token(self, audience="pbi") -> str:
        """Returns the bearer token for the environment (edog/dxt/msit)."""

        if audience not in ["pbi", "storage"]:
            raise ValueError(f"audience must be either 'pbi' or 'storage': {audience}")

        # useful to test using a specific token
        # import base64
        # token = ""
        # bearer_token_fabric = base64.b64decode(token.encode('ascii')).decode('ascii')
        # return bearer_token_fabric

        # check our local cache
        token = self.env_bearer_token.get(audience)
        if token is not None:
            return token

        password = self.get_env_password()
        if password is None:
            # usually from dev box
            from msal import PublicClientApplication

            app = PublicClientApplication(
                # PowerBI
                "7f67af8a-fedc-4b08-8b4e-37c4d127b6cf",
                authority="https://login.microsoftonline.com/organizations",
            )

            for aud in ["pbi", "storage"]:
                if aud == "pbi":
                    scope = "https://analysis.windows.net/powerbi/api/.default"
                else:
                    scope = "https://storage.azure.com/.default"

                self.env_bearer_token[aud] = app.acquire_token_interactive(
                    scopes=[scope], port=12604
                )["access_token"]
        else:
            # usually from CI pipeline. Password is fetched from KeyVault via ADO task + ADO service connection
            username = self.get_env_username()
            # useful to retrieve username/password to interactively explore your workspace
            print(
                f"ENV: {self.env} USERNAME: {username} PASSWORD: {hashlib.md5(password.encode('utf-8')).hexdigest()}"
            )

            for aud in ["pbi", "storage"]:
                self.env_bearer_token[aud] = get_bearer_token(
                    self.env, username, password, aud
                )

        return self.env_bearer_token[audience]

    def _get_or_create_workspace(self, workspace_name: str):
        from sempy.fabric._client._workspace_client import WorkspaceClient
        from sempy.fabric.exceptions._exceptions import WorkspaceNotFoundException

        try:
            workspace_id = WorkspaceClient(workspace_name).get_workspace_id()
        except WorkspaceNotFoundException:
            # get Premium Per User (PPU) capacity
            # on MSIT this one allows for XMLA edit
            # alternative is FT1
            import sempy.fabric as fabric
            from sempy.fabric import list_capacities

            capacity_id = (
                list_capacities().query("Sku == 'FT1' or Sku == 'F64'")["Id"].values[0]
            )

            workspace_id = fabric.create_workspace(
                workspace_name, capacity_id=capacity_id
            )

        return workspace_id

    def _get_workspace_name(self):
        alias = self.get_dev_user().split("@")[0]
        ws = f"Assistant {alias}"

        # allow external override
        return os.environ.get("ut_workspace", ws)

    def get_fabric_config(self) -> dict[str, str]:
        from sempy.fabric._token_provider import (
            SynapseTokenProvider,
            _get_token_seconds_remaining,
        )

        cache_path = ".fabric-config"

        if os.path.exists(cache_path):
            fabric_config = json.loads(open(cache_path, "r").read())

            # rehydrate bearer token, so we can resolve the user name
            self.env_bearer_token = {
                "pbi": fabric_config["bearer_token"],
                "storage": fabric_config["bearer_storage_token"],
            }

            workspace_name = self._get_workspace_name()

            # check if token is still valid and if the workspace/identity matches
            # in case user was switching between environments/identities:
            token_seconds_remaining = _get_token_seconds_remaining(
                fabric_config["bearer_token"]
            )
            cached_workspace_name = fabric_config.get("workspace_name", None)

            if (
                token_seconds_remaining > 10 * 60
                and cached_workspace_name == workspace_name
            ):
                return fabric_config
            else:
                # delete expired or wrong token
                self.env_bearer_token = dict()

        # this will trigger auth locally as we need the bearer token
        workspace_name = self._get_workspace_name()

        # get token and patch SynapseTokenProvider so flat api works
        bearer_token_fabric = self.get_env_bearer_token()
        setattr(SynapseTokenProvider, "__call__", lambda _: bearer_token_fabric)

        workspace_id = self._get_or_create_workspace(workspace_name)

        print(f"WORKSPACE: {workspace_name} ({workspace_id})")

        fabric_config = {
            "bearer_token": bearer_token_fabric,
            "bearer_storage_token": self.get_env_bearer_token("storage"),
            "workspace_id": workspace_id,
            "workspace_name": workspace_name,
            "username": self.get_dev_user(),
        }

        # cache the token
        with open(cache_path, "w") as f:
            f.write(json.dumps(fabric_config, default=str))

        return fabric_config


def get_fabric_info():
    # need to disable telemetry to avoid triggering all of SynapseML-Utils initialization
    os.environ["REPORT_USAGE_TELEMETRY"] = "false"

    from sempy.fabric._environment import _on_fabric

    if _on_fabric():
        import sempy.fabric as fabric

        from notebookutils.credentials import (  # pyright: ignore[reportMissingImports]
            getToken,
        )

        bearer_token = getToken("pbi")

        workspace_id = fabric.get_workspace_id()
        workspace_name = fabric.resolve_workspace_name(workspace_id)

        fabric_info = {
            "bearer_token": bearer_token,
            "workspace_id": workspace_id,
            "workspace_name": workspace_name,
        }
    else:
        from sempy.fabric import _environment, list_workspaces

        # default to MSIT to support local box development
        # the build pipeline overrides w/ the correct environment
        env = os.environ.get("ut_env", "daily")
        _environment.environment = env

        fabric_info = FabricAuthUtil(env).get_fabric_config()

        workspace_id = fabric_info["workspace_id"]

        # still need to patch sempy environment
        def _local_get_trident_config(key: str) -> str:
            if key == "trident.workspace.id" or key == "trident.artifact.workspace.id":
                return fabric_info["workspace_id"]
            elif key == "trident.onelake.endpoint":
                return env_onelake_new[env]
            else:
                raise ValueError(f"Unsupported key '{key}'")

        setattr(_environment, "_get_trident_config", _local_get_trident_config)

        # must not import to early due to how things are setup
        from synapse.ml.internal_utils.session_utils import (
            set_fabric_context,
            get_fabric_context,
        )
        from synapse.ml.fabric.service_discovery import set_envs
        from synapse.ml.fabric.token_utils import TokenUtils

        # TODO: AI Skill naming to revisit later
        ctx = {
            "trident.aiskill.shared_host": env_workload_host[env],
            "trident.lakehouse.tokenservice.endpoint": f"https://{env_workload_host[env]}",
            # "trident.capacity.id": test_environment.capacity_id,
            "trident.artifact.workspace.id": workspace_id,
            # for some reason sempy is using that key to lookup the workspace id
            "currentWorkspaceId": workspace_id,
            # "trident.lakehouse.id": lakehouse_id,
            "trident.workspace.id": workspace_id,
            # "trident.artifact.id": test_environment.artifact_id,
            "spark.trident.pbienv": env,
            "trident.aiskill.env": "true",
            "fs.defaultFS": f"https://{env_onelake_new[env]}",
        }

        set_fabric_context(ctx)

        # 2. Set AAD token
        TokenUtils().AAD_TOKEN = fabric_info["bearer_token"]
        TokenUtils().ML_AAD_TOKEN = fabric_info["bearer_token"]

        # 3. Set envs
        set_envs()

        # automatically resolve capacity id, can only do it after the environment is setup so we can call Fabric
        capacity_id = list_workspaces(filter=f"id eq '{workspace_id}'").iloc[0][
            "Capacity Id"
        ]
        ctx["trident.capacity.id"] = capacity_id

        # need to update the context again
        set_fabric_context(ctx)
        set_envs()

    return fabric_info
